from copy import deepcopy
from typing import List, Tuple

from PyQt5.QtCore import pyqtSlot, pyqtSignal, QObject

from .roi import Roi, RoiTypes
from .roi_data import RoiData
from ..file_manager import FileManager, ImageKey
from ..geometry_holder import GeometryHolder


def _check_non_empty(func):
    def wrapper(self, *args, **kwargs):
        if not self._current_key:
            return
        else:
            return func(self, *args, **kwargs)

    return wrapper


class RoiDict(QObject):
    sig_roi_created = pyqtSignal(tuple)
    sig_roi_deleted = pyqtSignal(tuple)
    sig_roi_moved = pyqtSignal(tuple, str)

    sig_selected = pyqtSignal(tuple)
    sig_one_selected = pyqtSignal(int)
    sig_roi_renamed = pyqtSignal(int)
    sig_fixed = pyqtSignal(tuple)
    sig_unfixed = pyqtSignal(tuple)

    sig_type_changed = pyqtSignal(int)
    sigFitRoisOpen = pyqtSignal(list)

    EMIT_NAME = 'RoiDict'

    def __init__(self, file_manager: FileManager, geometry_holder: GeometryHolder):
        super().__init__()
        self._geometry_holder: GeometryHolder = geometry_holder
        self._fm: FileManager = file_manager
        self._roi_data: RoiData = RoiData()
        self._current_key: ImageKey or None = None
        self._copied_rois: CopiedRois = CopiedRois()

    @property
    def is_copied(self):
        return len(self._copied_rois) > 0

    @property
    def ring_bounds(self) -> Tuple[float, float]:
        return self._geometry_holder.geometry.ring_bounds

    @_check_non_empty
    def save_state(self):
        if len(self._roi_data):
            self._fm.rois_data[self._current_key] = self._roi_data
        else:
            del self._fm.rois_data[self._current_key]

    def clear(self):
        if len(self._roi_data):
            self.sig_roi_deleted.emit(tuple(self.keys()))
            self._roi_data.clear()

    def save_and_clear(self):
        self.save_state()
        self.clear()

    def change_image(self, image_key: ImageKey):
        self.clear()
        self._current_key = image_key
        if not self._current_key:
            return
        self._update()

    # def _update_real_time(self):
    #
    #     roi_data = self._fm.rois_data[self._current_key] or RoiData()
    #     prev_key = self._current_key.get_previous()
    #     if prev_key:
    #         prev_data = self._fm.rois_data[prev_key] or RoiData()
    #     else:
    #         prev_data = RoiData()
    #
    #     self._merge_data(roi_data, prev_data)

    def _update(self):
        self._roi_data = self._fm.rois_data[self._current_key] or RoiData()
        self._roi_data.change_ring_bounds(self._geometry_holder.geometry.ring_bounds)
        if len(self._roi_data):
            self.sig_roi_created.emit(tuple(self.keys()))

    def _merge_data(self, roi_data: RoiData, prev_data: RoiData = None):
        prev_data = prev_data or {}

        to_move = []
        to_create = []
        to_delete = []

        def _add_roi(roi_):
            key_ = roi_.key
            if key_ in self.keys():
                prev_roi = self[key_]
                if prev_roi.deleted and roi_.deleted:
                    pass
                elif prev_roi.deleted and not roi_.deleted:
                    to_create.append(key_)
                elif not prev_roi.deleted and not roi_.deleted:
                    to_move.append(key_)
                elif not prev_roi.deleted and roi_.deleted:
                    to_delete.append(key_)
                roi_.active = prev_roi.active
                prev_roi.update(roi_)

            else:
                self._roi_data.add_roi(roi_)
                if not roi_.deleted:
                    to_create.append(key_)

        for key in set(roi_data.keys()).union(prev_data.keys()).union(self.keys()):
            if key in roi_data:
                _add_roi(roi_data[key])
            elif key in prev_data:
                _add_roi(prev_data[key])
            else:
                roi = self[key]
                if not roi.deleted:
                    roi.deleted = True
                    to_delete.append(key)

        if to_move:
            self.sig_roi_moved.emit(tuple(to_move), self.EMIT_NAME)
        if to_create:
            self.sig_roi_created.emit(tuple(to_create))
        if to_delete:
            self.sig_roi_deleted.emit(tuple(to_delete))

    @property
    def selected_rois(self) -> List[Roi]:
        return self._roi_data.selected_rois

    def keys(self):
        yield from self._roi_data.keys()

    def values(self):
        yield from self._roi_data.values()

    def items(self):
        yield from self._roi_data.items()

    def __getitem__(self, item: int) -> Roi:
        return self._roi_data[item]

    @pyqtSlot(int, name='select')
    def select(self, key: int):
        change_select = self._roi_data.select(key)
        if change_select:
            self.sig_selected.emit(change_select)
            if self._roi_data.selected_num == 1:
                self.sig_one_selected.emit(next(iter(self._roi_data.selected_keys)))

    @pyqtSlot(int, name='shift_select')
    def shift_select(self, key: int):
        self._roi_data.shift_select(key)
        self.sig_selected.emit((key,))
        if self._roi_data.selected_num == 1:
            self.sig_one_selected.emit(next(iter(self._roi_data.selected_keys)))

    @pyqtSlot(name='selectAll')
    @_check_non_empty
    def select_all(self) -> None:
        self._emit_select(self._roi_data.select_all())

    @pyqtSlot(name='unselectAll')
    @_check_non_empty
    def unselect_all(self):
        self._emit_select(self._roi_data.unselect_all())

    @pyqtSlot(name='deleteSelected')
    @_check_non_empty
    def delete_selected_roi(self):
        for key in list(self._roi_data.selected_keys):
            self.delete_roi(key)

    @pyqtSlot(int, str, name='changeName')
    def change_name(self, key: int, name: str):
        try:
            self[key].name = name
            self.sig_roi_renamed.emit(key)
        except KeyError:
            return

    def _emit_select(self, keys: List[int]):
        if keys:
            self.sig_selected.emit(tuple(keys))
        if self._roi_data.selected_num == 1:
            self.sig_one_selected.emit(next(iter(self._roi_data.selected_keys)))

    def add_roi(self, roi: Roi) -> None:
        self._roi_data.add_roi(roi)
        self.sig_roi_created.emit((roi.key,))
        if roi.active:
            self._emit_select((roi.key,))

    def add_rois(self, roi_list: List[Roi]):
        to_select = self._roi_data.add_rois(roi_list)
        self.sig_roi_created.emit(tuple(roi.key for roi in roi_list))
        self._emit_select(to_select)

    @_check_non_empty
    def delete_roi(self, key: int):
        self._roi_data.delete_roi(key)
        self.sig_roi_deleted.emit((key,))

    @_check_non_empty
    def create_roi(self, radius: float = None, width: float = None, **params) -> Roi:
        d = self._default_params()
        d.update(params)
        if radius and width:
            d.update(radius=radius, width=width)
        roi = self._roi_data.create_roi(**d)
        self.sig_roi_created.emit((roi.key,))
        if roi.active:
            self._emit_select((roi.key,))
        return roi

    @pyqtSlot(tuple, name='change_ring_bounds')
    def change_ring_bounds(self, bounds: Tuple[float, float]):
        keys = self._roi_data.change_ring_bounds(bounds)
        if keys:
            self.sig_roi_moved.emit(tuple(keys), self.EMIT_NAME)

    @pyqtSlot(name='on_scale_changed')
    def on_scale_changed(self):
        self._roi_data.on_scale_changed(self._geometry_holder.geometry.scale_change)
        self.sig_roi_moved.emit(tuple(self.keys()), self.EMIT_NAME)

    def _default_params(self) -> dict:
        r0, r1 = self._geometry_holder.geometry.r_range
        r0 *= self._geometry_holder.geometry.scale
        r1 *= self._geometry_holder.geometry.scale
        radius = (r0 + r1) / 2
        width = (r1 - r0) / 10
        angle, angle_std = self.ring_bounds
        return dict(radius=radius, width=width, angle=angle, angle_std=angle_std)

    @pyqtSlot(int, str, name='moveRoi')
    def move_roi(self, key: int, name: str):
        self.select(key)
        self.sig_roi_moved.emit((key,), name)
        angle, angle_std = self.ring_bounds
        roi = self[key]
        if roi.type == RoiTypes.ring and (roi.angle != angle or roi.angle_std != angle_std):
            roi.type = RoiTypes.segment
            self.sig_type_changed.emit(key)

    @pyqtSlot(int, name='roiTypeChanged')
    def change_roi_type(self, key: int):
        try:
            roi = self[key]
            if roi.type == RoiTypes.ring and (roi.angle, roi.angle_std) != self.ring_bounds:
                roi.angle, roi.angle_std = self.ring_bounds
                self.sig_roi_moved.emit((key,), self.EMIT_NAME)
        except KeyError:
            return
        self.sig_type_changed.emit(key)

    @pyqtSlot(name='fixAll')
    def fix_all(self, only_selected: bool = False):
        keys = []
        for roi in self.values():
            if only_selected and not roi.active:
                continue
            if roi.movable:
                roi.movable = False
                keys.append(roi.key)
        if keys:
            self.sig_fixed.emit(tuple(keys))

    @pyqtSlot(name='fixSelected')
    def fix_selected(self):
        self.fix_all(True)

    @pyqtSlot(name='unfixAll')
    def unfix_all(self, only_selected: bool = False):
        keys = []
        for roi in self.values():
            if only_selected and not roi.active:
                continue
            if not roi.movable:
                roi.movable = True
                keys.append(roi.key)
        if keys:
            self.sig_unfixed.emit(tuple(keys))

    @pyqtSlot(name='unfixSelected')
    def unfix_selected(self):
        self.unfix_all(True)

    @pyqtSlot(int, name='fixRoi')
    def fix_roi(self, key: int):
        try:
            self[key].movable = False
            self.sig_fixed.emit((key,))
        except KeyError:
            return

    @pyqtSlot(int, name='unfixRoi')
    def unfix_roi(self, key: int):
        try:
            self[key].movable = True
            self.sig_unfixed.emit((key,))
        except KeyError:
            return

    def copy_rois(self, key: int or str = 'selected'):
        if key == 'selected':
            rois = self.selected_rois
        elif key == 'all':
            rois = list(self.values())
        else:
            try:
                rois = [self[key]]
            except KeyError:
                return
        if rois:
            self._copied_rois.copy_rois(rois, self._geometry_holder.geometry)

    def paste_rois(self):
        if not self._current_key or not len(self._copied_rois):
            return
        self.add_rois(self._copied_rois.paste(self._geometry_holder.geometry))

    @pyqtSlot(bool, name='openFitRois')
    def open_fit_rois(self, only_selected: bool):
        if only_selected:
            rois = deepcopy(self.selected_rois)
        else:
            rois = deepcopy(list(self.values()))
        if rois:
            self.sigFitRoisOpen.emit(rois)

    def apply_fit(self, rois: List[Roi], image_key: ImageKey):

        if self._current_key == image_key:
            keys_to_create, keys_to_move = self._roi_data.apply_fit(rois)
            self.sig_roi_created.emit(keys_to_create)
            self.sig_roi_moved.emit(keys_to_move, self.EMIT_NAME)
            self.sig_fixed.emit(keys_to_move)
            self.sig_selected.emit(keys_to_move)

        else:
            roi_data = self._fm.rois_data[image_key] or RoiData()
            roi_data.apply_fit(rois)
            self._fm.rois_data[image_key] = roi_data


class CopiedRois(object):
    def __init__(self, rois: List[Roi] = None, geometry=None):
        self._rois = None
        self._scale = None
        self._bounds = None

        if rois and geometry:
            self.copy_rois(rois, geometry)

    def __len__(self):
        return len(self._rois) if self._rois else 0

    def copy_rois(self, rois: List[Roi], geometry):
        self._rois = deepcopy(rois)

        for roi in self._rois:
            roi.active = True

        self._scale = geometry.scale
        self._bounds = geometry.ring_bounds

    def paste(self, geometry, *, clear_keys: bool = True,
              clear: bool = False):
        if not self._rois:
            return

        if clear:
            rois = self._rois
            self.clear()
        else:
            rois = deepcopy(self._rois)

        if self._scale != geometry.scale:
            s = geometry.scale / self._scale
            for roi in rois:
                roi.radius *= s
                roi.width *= s
        if self._bounds != geometry.ring_bounds:
            for roi in rois:
                if roi.type == RoiTypes.ring:
                    roi.angle, roi.angle_std = geometry.ring_bounds
        if clear_keys:
            for roi in rois:
                roi.key = None

        return rois

    def clear(self):
        self._rois = None
        self._scale = None
        self._bounds = None



# -*- coding: utf-8 -*-
from dataclasses import dataclass

import numpy as np

from ..file_manager import FileManager, ImageKey
from .basic_profile import BasicProfile, BaselineParams


@dataclass
class SavedProfile:
    data: np.ndarray
    sigma: float
    baseline_params: BaselineParams
    baseline: np.ndarray = None


class RadialProfile(BasicProfile):
    def __init__(self, image_holder, fm: FileManager, parent=None):
        self.image_holder = image_holder
        self.fm = fm
        self._current_key: ImageKey = None
        super().__init__(parent)

    def save_state(self):
        if self._current_key and self._raw_y is not None:
            self.fm.profiles[self._current_key] = self.to_save()

    def update_data_from_source(self):
        # self.save_state()

        self._current_key = self.image_holder.current_key
        r_axis = self.image_holder.geometry.r_axis

        if r_axis is None:
            return

        saved_profile = self.fm.profiles[self._current_key] if self._current_key else None

        if not saved_profile or np.any(saved_profile.x != r_axis):
            profile = self.image_holder.get_radial_profile()
            if profile is None:
                return
            self.set_data(profile, r_axis)
        else:
            self.from_save(saved_profile)



from typing import Dict
import numpy as np
from copy import deepcopy

from ..file_manager import ImageKey
from ..profiles import SavedProfile
from ..utils import smooth_curve, baseline_correction
from .background import *
from .functions import *
from .fit import Fit
from .utils import _get_dummy_bounds, Roi, RoiTypes
from .range_strategy import RangeStrategy, RangeStrategyType


class FitObject(object):
    MINIMAL_NUM: int = 5

    def __init__(self, image_key: ImageKey, polar_image: np.ndarray,
                 r_axis: np.ndarray, phi_axis: np.ndarray, *, saved_profile: SavedProfile = None,
                 update_baseline: bool = False):
        self.image_key = image_key

        self.polar_image = polar_image
        self.r_axis = r_axis
        self.phi_axis = phi_axis
        self.r_delta = (r_axis.max() - r_axis.min()) / r_axis.size
        self.min_range: float = self.r_delta * self.MINIMAL_NUM
        self.phi_delta = (phi_axis.max() - phi_axis.min()) / phi_axis.size
        self.r_profile = polar_image.sum(axis=0)
        self.aspect_ratio = self._aspect_ratio()
        self.bounds = self._bounds()

        self.fits: Dict[int, Fit] = {}
        self.name: str = ''
        self.is_fitted: bool = False
        self.saved_profile: SavedProfile or None = None

        self.default_fitting: FittingFunction.__class__ = Gaussian
        self.default_background: Background.__class__ = LinearBackground
        self.default_range_strategy: RangeStrategy = RangeStrategy()

        if saved_profile:
            self.set_profile(saved_profile, update_baseline)

    def set_profile(self, saved_profile: SavedProfile, update_baseline: bool = False):
        if np.any(saved_profile.x != self.r_axis):
            return

        self.saved_profile = saved_profile

        if update_baseline:
            self.update_baseline()
        else:
            self.r_profile = smooth_curve(saved_profile.raw_data, saved_profile.sigma)
            if saved_profile.baseline is not None:
                self.r_profile = self.r_profile - saved_profile.baseline

        self.update_fit_data(update_r_range=False, update_fit=True)

    def update_baseline(self):
        if self.saved_profile:
            self.saved_profile.x = self.r_axis
            r1, r2 = self.saved_profile.x_range
            x1, x2 = self._get_r_coords(r1), self._get_r_coords(r2)
            self.saved_profile.raw_data = self.polar_image.sum(axis=0)
            self.r_profile = smooth_curve(self.saved_profile.raw_data, self.saved_profile.sigma)

            baseline = np.zeros_like(self.r_profile)

            baseline[x1:x2] = baseline_correction(
                self.r_profile[x1:x2], self.saved_profile.baseline_params.smoothness,
                self.saved_profile.baseline_params.asymmetry)

            self.saved_profile.baseline = baseline
            self.r_profile = self.r_profile - baseline
            self.update_fit_data(update_r_range=False, update_fit=True)

    def clear_profile(self):
        self.saved_profile = None
        self.r_profile = self.polar_image.sum(axis=0)

    def add_fit(self, fit: Fit):
        self.fits[fit.roi.key] = fit
        update_r_range = fit.range_strategy.strategy_type == RangeStrategyType.adjust
        self.update_fit_data(fit, update_r_range, update_fit=True)

    def new_fit(self, roi: Roi):
        r1, r2 = self._get_r_range(roi, self.default_range_strategy.range_factor)

        x1, x2 = self._get_r_coords(r1), self._get_r_coords(r2)
        x, y = self._get_x_y(roi, x1, x2)

        background: Background = self.default_background()
        function: FittingFunction = self.default_fitting()

        if x.size:
            init_params, upper_bounds, lower_bounds = function.bounds(x, y, roi, background, params_from_roi=True)
        else:
            init_params, upper_bounds, lower_bounds = _get_dummy_bounds(function.NUM + background.NUM)

        background_curve = background(x, *init_params)
        init_curve = function(x, background, *init_params)

        fit = Fit(roi=roi, r_range=(r1, r2), x_range=(x1, x2),
                  x=x, y=y, init_curve=init_curve, init_params=init_params,
                  lower_bounds=lower_bounds, upper_bounds=upper_bounds,
                  fitted_params=[], fit_errors=[], fitting_curve=None,
                  fitting_function=function, background=background,
                  background_curve=background_curve, range_strategy=deepcopy(self.default_range_strategy))

        self.fits[roi.key] = fit

        return fit

    def update_fit_data(self, fit: Fit = None, update_r_range: bool = True, *, update_fit: bool = False, **kwargs):
        if fit:
            self._update_fit_data(fit, update_r_range, update_fit=update_fit, **kwargs)
        else:
            for fit in self.fits.values():
                self._update_fit_data(fit, update_r_range, update_fit=update_fit, **kwargs)

    def _update_fit_data(self, fit: Fit, update_r_range: bool = True, *, update_fit: bool = False, **kwargs):
        if update_r_range:
            fit.r_range = r1, r2 = self._get_r_range(fit.roi, fit.range_strategy.range_factor)
        else:
            r1, r2 = fit.r_range
        x1, x2 = self._get_r_coords(r1), self._get_r_coords(r2)
        fit.x, fit.y = self._get_x_y(fit.roi, x1, x2)

        if update_fit:
            fit.update_fit(**kwargs)

    def remove_fit(self, fit: Fit):
        try:
            del self.fits[fit.roi.key]
        except KeyError:
            return

    def set_range_strategy(self, range_strategy: RangeStrategy, fit: Fit = None, update: bool = True):
        if fit:
            range_strategy.is_default = False
            fit.range_strategy = range_strategy
        else:
            self.default_range_strategy = range_strategy
            for fit in self.fits.values():
                if fit.range_strategy.is_default:
                    fit.range_strategy = deepcopy(range_strategy)
                    if update:
                        self.update_fit_data(fit, update_fit=True)

    def set_background(self, background_type: BackgroundType, fit: Fit = None, update: bool = True):
        background = BACKGROUNDS[background_type]
        if fit and fit.background.TYPE != background_type:
            fit.set_background(background())
            fit.background.is_default = False
        elif self.default_background.TYPE != background_type:
            self.default_background = background
            for fit in self.fits.values():
                if fit.background.is_default:
                    fit.set_background(background(), update)

    def set_function(self, fitting_type: FittingType, fit: Fit = None, update: bool = True):
        fitting_function = FITTING_FUNCTIONS[fitting_type]
        if fit and fit.fitting_function.TYPE != fitting_function:
            fit.set_function(fitting_function())
            fit.fitting_function.is_default = False
        elif self.default_fitting.TYPE != fitting_type:
            self.default_fitting = fitting_function
            for fit in self.fits.values():
                if fit.fitting_function.is_default:
                    fit.set_function(fitting_function(), update)

    # Private methods for calculating geometric properties

    def _get_r_range(self, roi: Roi, factor: float):
        r1, r2 = roi.radius - roi.width * (factor + 1), roi.radius + roi.width * (factor + 1)
        return min(r1, roi.radius - self.min_range / 2), max(r2, roi.radius + self.min_range / 2)

    def _get_x_y(self, roi: Roi, x1: int, x2: int):
        x = self.r_axis[x1:x2]

        if roi.type == RoiTypes.ring:
            y = self.r_profile[x1:x2]
        else:
            p1, p2 = self._get_p_coords(roi.angle - roi.angle_std / 2), \
                     self._get_p_coords(roi.angle + roi.angle_std / 2)
            y = self.polar_image[max(0, p1):min(p2, self.phi_axis.size - 1), x1:x2].sum(axis=0)

        return x, y

    def _get_r_coords(self, r):
        return int((r - self.r_axis.min()) / self.r_delta)

    def _get_p_coords(self, p):
        return int((p - self.phi_axis.min()) / self.phi_delta)

    def _aspect_ratio(self):
        p, r = self.phi_axis, self.r_axis
        return (p.max() - p.min()) * p.size / (r.max() - r.min()) / r.size

    def _bounds(self):
        p = self.phi_axis
        return (p.max() + p.min()) / 2, (p.max() - p.min())



# -*- coding: utf-8 -*-

import numpy as np

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidgetAction

from pyqtgraph import (GraphicsLayoutWidget, setConfigOptions,
                       ImageItem, HistogramLUTItem)
from pyqtgraph.graphicsItems.ViewBox.ViewBoxMenu import ViewBoxMenu

from giwaxs_gui.gui.basic_widgets.sliders import LabeledSlider


class CustomViewBoxMenu(ViewBoxMenu):
    sigSigmaChanged = pyqtSignal(float)
    sigRangeAsDefault = pyqtSignal()

    def __init__(self, view, sigma: float = 3):
        super().__init__(view)
        self.slider = LabeledSlider('Clip sigma factor', bounds=(0.001, 4),
                                    value=sigma, parent=self)
        self.slider.valueChanged.connect(self.sigSigmaChanged.emit)
        self.addAction('Set range as default', lambda *x: self.sigRangeAsDefault.emit())
        self.sigma_factor_menu = self.addMenu('Set sigma factor')
        action = QWidgetAction(self)

        action.setDefaultWidget(self.slider)
        self.sigma_factor_menu.addAction(action)


class CustomImageViewer(GraphicsLayoutWidget):
    @property
    def view_box(self):
        return self.image_plot.vb

    def __init__(self, parent=None, *, hist_range: tuple = None, sigma_factor: float = 3, **kwargs):
        setConfigOptions(imageAxisOrder='row-major')
        super(CustomImageViewer, self).__init__(parent)
        self._scale = (1., 1.)
        self._center = (0, 0)

        self._hist_range = hist_range
        self._sigma_factor = sigma_factor

        self._init_ui(**kwargs)

    def _init_ui(self, **kwargs):
        self.setWindowTitle('Image Viewer')
        self.image_plot = self.addPlot(**kwargs)
        self.image_plot.vb.setAspectLocked()
        self.image_plot.vb.invertY()
        self.image_item = ImageItem()
        self.image_plot.addItem(self.image_item)
        self.image_plot.setMenuEnabled(False)
        self.hist = HistogramLUTItem()
        self.hist.setImageItem(self.image_item)
        self.addItem(self.hist)
        self.hist.vb.menu = CustomViewBoxMenu(self.hist.vb)
        self.hist.vb.menu.sigSigmaChanged.connect(self.set_sigma_factor)
        self.hist.vb.menu.sigRangeAsDefault.connect(self.set_limit_as_default)

    def set_data(self, data, *, reset_axes: bool = False):
        if data is None:
            return
        self.image_item.setImage(data)
        self.set_levels()
        if reset_axes:
            self.image_item.resetTransform()
        self.set_default_range()

    def hist_params(self) -> dict:
        return dict(sigma_factor=self._sigma_factor, hist_range=self._hist_range)

    def clear_image(self):
        self.set_data(np.zeros((1, 1)))

    def set_default_range(self):
        if self.image_item.image is None:
            return
        # self.set_auto_range()
        axes = self.get_axes()
        self.image_plot.setRange(xRange=axes[1], yRange=axes[0])

    def set_auto_range(self):
        self.image_plot.autoRange()

    def set_levels(self):
        img = self.image_item.image
        if img is None:
            return

        if self._sigma_factor and self._sigma_factor > 0:
            m, s = img.flatten().mean(), img.flatten().std() * self._sigma_factor
            self.hist.setLevels(max(m - s, img.min()), min(m + s, img.max()))
        elif self._hist_range:
            self.hist.setLevels(*self._hist_range)
        else:
            self.hist.setLevels(self.image_item.image.min(),
                                self.image_item.image.max())

    def set_sigma_factor(self, sigma_factor: float):
        self._sigma_factor = sigma_factor
        self.set_levels()

    def set_limit_as_default(self):
        self._hist_range = self.hist.getLevels()
        self._sigma_factor = None
        self.set_levels()

    def get_levels(self):
        return self.hist.getLevels()

    def set_center(self, center: tuple, pixel_units: bool = True):
        if not pixel_units:
            scale = self.get_scale()
            center = (center[0] / scale[0], center[1] / scale[1])
        if self._center != (0, 0) or self._scale != (1., 1.):
            self.image_item.resetTransform()
            self.image_item.scale(*self._scale)
        self.image_item.translate(- center[0], - center[1])
        self._center = center
        self.set_default_range()

    def set_scale(self, scale: float or tuple):
        if isinstance(scale, float) or isinstance(scale, int):
            scale = (scale, scale)
        if self._center != (0, 0) or self._scale != (1., 1.):
            self.image_item.resetTransform()
        self.image_item.scale(*scale)
        if self._center != (0, 0):
            self.image_item.translate(- self._center[0], - self._center[1])
        self._scale = scale
        self.set_default_range()

    def get_scale(self):
        # scale property is occupied by Qt superclass.
        return self._scale

    def get_center(self):
        return self._center

    def set_x_axis(self, x_min, x_max):
        self._set_axis(x_min, x_max, 0)
        self.set_default_range()

    def set_y_axis(self, y_min, y_max):
        self._set_axis(y_min, y_max, 1)
        self.set_default_range()

    def _set_axis(self, min_: float, max_: float, axis_ind: int):
        shape = self.image_item.image.shape
        scale = np.array(self._scale)
        scale[axis_ind] = (max_ - min_) / shape[axis_ind]
        center = np.array(self._center)
        center[axis_ind] = - min_ / scale[axis_ind]
        if self._center != (0, 0) or self._scale != (1., 1.):
            self.image_item.resetTransform()
        self.image_item.scale(scale[0], scale[1])
        self.image_item.translate(- center[0], - center[1])
        self._scale = tuple(scale)
        self._center = tuple(center)

    def get_axes(self):
        shape = np.array(self.image_item.image.shape)
        scale = np.array(self._scale)
        min_ = - np.array((self._center[1], self._center[0])) * scale
        max_ = min_ + shape * scale
        return (min_[0], max_[0]), (min_[1], max_[1])


if __name__ == '__main__':
    import sys
    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)
    window = CustomImageViewer()
    image = np.random.randint(0, 100, (100, 150))
    window.set_data(image)
    window.set_center((100, 0))
    window.set_scale(0.01)
    window.show()
    sys.exit(app.exec_())



from pathlib import Path

from PyQt5.QtWidgets import (QTreeView, QMenu,
                             QWidget, QHBoxLayout, QLabel)
from PyQt5.QtGui import QStandardItem, QStandardItemModel
from PyQt5.QtCore import Qt

from ..basic_widgets import RoundedPushButton
from ..tools import Icon, get_folder_filepath, get_image_filepath
from ...app.file_manager import FileManager, ImageKey, FolderKey


class FileModel(QStandardItemModel):
    def __init__(self):
        super(FileModel, self).__init__()
        self.setHorizontalHeaderLabels([''])
        self.setRowCount(0)

    def new_project(self):
        self.removeRows(1, self.rowCount() - 1)


class ImageItem(QStandardItem):

    def __init__(self, key: ImageKey):
        super().__init__(key.name)
        self.key: ImageKey = key
        self.setIcon(Icon('data'))


class FolderItem(QStandardItem):

    def __init__(self, key: FolderKey):
        super().__init__(key.name)
        self.key = key
        self.setIcon(Icon('folder'))
        self._updated: bool = False

    def on_clicked(self):
        if not self._updated:
            self.update()
        self._updated = True

    def _fill(self):
        for folder in self.key.folder_children:
            self.appendRow(FolderItem(folder))
        for image in self.key.image_children:
            self.appendRow(ImageItem(image))

    def update(self):
        self.clear()
        self.key.update()
        self._fill()

    def clear(self):
        self.removeRows(0, self.rowCount())


class FileViewer(QTreeView):

    def __init__(self, fm: FileManager, parent=None):
        super().__init__(parent)
        self._model = FileModel()
        self.setModel(self._model)
        self.setEditTriggers(QTreeView.NoEditTriggers)
        self._fm = fm
        self._fm.sigNewFile.connect(self._add_new_file)
        self._fm.sigNewFolder.connect(self._add_new_folder)
        self._fm.sigProjectClosed.connect(self._model.new_project)

        self.selectionModel().currentChanged.connect(self._on_clicked)
        self.customContextMenuRequested.connect(self._context_menu)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self._init_ui()
        self.show()

    def _init_ui(self):
        add_file_button = RoundedPushButton(icon=Icon('data'), radius=30,
                                            background_color='transparent')
        add_file_button.clicked.connect(self._open_add_file_menu)
        add_folder_button = RoundedPushButton(icon=Icon('folder'), radius=30,
                                              background_color='transparent')
        add_folder_button.clicked.connect(self._open_add_folder_menu)
        layout = self._get_header_layout(QStandardItem(), 'Files')
        layout.addWidget(add_file_button)
        layout.addWidget(add_folder_button)

    def _get_header_layout(self, item: QStandardItem, label: str):
        header_widget = QWidget(self)
        layout = QHBoxLayout()
        header_widget.setLayout(layout)
        label_widget = QLabel(label)
        layout.addWidget(label_widget, alignment=Qt.AlignLeft)
        layout.addStretch(1)
        self._model.appendRow(item)
        self.setIndexWidget(item.index(), header_widget)
        return layout

    def _add_new_file(self, key: ImageKey):
        self._model.appendRow(ImageItem(key))

    def _add_new_folder(self, key: FolderKey):
        self._model.appendRow(FolderItem(key))

    def _on_clicked(self, index):
        item = self._model.itemFromIndex(index)
        if isinstance(item, ImageItem):
            self._fm.change_image(item.key)
        elif isinstance(item, FolderItem):
            item.on_clicked()
            self.setExpanded(item.index(), True)

    def _open_add_file_menu(self):
        path = get_image_filepath(self)
        if path:
            self._fm.add_root_path_to_project(path)

    def _open_add_folder_menu(self):
        path = get_folder_filepath(self, message='Choose directory containing images or h5 files')
        if path:
            self._fm.add_root_path_to_project(path)

    def _context_menu(self, position):
        item = self._model.itemFromIndex(self.indexAt(position))
        menu = QMenu()
        if isinstance(item, FolderItem):
            update_folder = menu.addAction('Update folder')
            update_folder.triggered.connect(item.update)
            close_folder = menu.addAction('Remove from project')
            close_folder.triggered.connect(
                lambda *x, it=item: self._remove_item(it))

        elif isinstance(item, ImageItem):
            close_image = menu.addAction('Remove from project')
            close_image.triggered.connect(
                lambda *x, it=item: self._remove_item(it))
        else:
            return
        menu.exec_(self.viewport().mapToGlobal(position))

    def _remove_item(self, item: FolderItem or ImageItem):
        parent = item.parent() or self._model
        parent.removeRow(item.row())
        self._fm.remove_key(item.key)



from h5py import Group

from .npy_file_manager import _ReadNpy


class _ReadImage(_ReadNpy):
    NAME = 'images'

    @staticmethod
    def _get_h5(h5group: Group, key):
        pass

    @staticmethod
    def _set_h5(h5group: Group, key, image):
        _ReadImage._del_h5(h5group, key)
        h5group.create_dataset('image', data=image)

    @staticmethod
    def _del_h5(h5group: Group, key):
        if 'image' in h5group.keys():
            del h5group['image']

    def __getitem__(self, key):
        return key.get_image()




# -*- coding: utf-8 -*-

from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot

from pyqtgraph import LinearRegionItem

from ...app import Roi, App

from .abstract_roi_widget import AbstractRoiWidget


class Roi1D(AbstractRoiWidget, LinearRegionItem):
    BRIGHT_COLOR: bool = False

    sigRoiMoved = pyqtSignal(int)
    sigSelected = pyqtSignal(int)
    sigShiftSelected = pyqtSignal(int)

    _ACTIVE_Z = 10
    _INACTIVE_Z = 5
    _FIXED_Z = -30

    def __init__(self, roi: Roi, enable_context: bool = True, **kwargs):
        AbstractRoiWidget.__init__(self, roi, enable_context=enable_context, **kwargs)
        LinearRegionItem.__init__(self)
        self.sigRegionChanged.connect(self.roi_is_moving)
        self.update_roi()

        if App().debug_tracker:
            App().debug_tracker.add_object(self, roi.name)

    def roi_is_moving(self):
        if self.moving:
            self.send_move()
        for l in self.lines:
            if l.moving:
                self.send_move()

    def set_color(self, color):
        self.setBrush(color)
        self.update()

    def set_movable(self, movable: bool):
        self.setMovable(movable)

    def send_move(self):
        x1, x2 = self.getRegion()
        r, w = (x1 + x2) / 2, x2 - x1

        self.roi.radius = r
        self.roi.width = w

        self.sigRoiMoved.emit(self.roi.key)

    @pyqtSlot(name='move_roi')
    def move_roi(self):
        r, w = self.roi.radius, abs(self.roi.width)
        x1, x2 = r - w / 2, r + w / 2
        self.setRegion((x1, x2))
        self.update()

    def hoverEvent(self, ev):
        pass

    def mouseDragEvent(self, ev):
        if not ev.modifiers() == Qt.ShiftModifier:
            super().mouseDragEvent(ev)
            self.send_move()
        else:
            self.mouseClickEvent(ev)

    def mouseClickEvent(self, ev):
        if self.moving and ev.button() == Qt.RightButton:
            ev.accept()
            for i, l in enumerate(self.lines):
                l.setPos(self.startPositions[i])
            self.moving = False
            self.sigRegionChanged.emit(self)
            self.sigRegionChangeFinished.emit(self)
            self.send_move()
        elif ev.button() == Qt.RightButton:
            ev.accept()
            self.show_context_menu(ev)
        elif ev.button() == Qt.LeftButton and ev.modifiers() == Qt.ShiftModifier:
            ev.accept()
            self.sigShiftSelected.emit(self.roi.key)

        elif ev.button() == Qt.LeftButton:
            ev.accept()
            self.sigSelected.emit(self.roi.key)
        else:
            ev.ignore()

        self.viewRangeChanged()

    def update_select(self):
        super().update_select()
        self._update_z_value()

    def fix(self):
        super().fix()
        self.setMovable(False)
        self._update_z_value()

    def unfix(self):
        super().unfix()
        self.setMovable(True)
        self._update_z_value()

    def _update_z_value(self):
        if self.roi.movable:
            if self.roi.active:
                self.setZValue(self._ACTIVE_Z)
            else:
                self.setZValue(self._INACTIVE_Z)
        else:
            self.setZValue(self._FIXED_Z)
        self.update()


class Roi1DAngular(Roi1D):

    @pyqtSlot(name='move_roi')
    def move_roi(self):
        a, a_w = self.roi.angle, self.roi.angle_std
        x1, x2 = a - a_w / 2, a + a_w / 2
        self.setRegion((x1, x2))

    def send_move(self):
        x1, x2 = self.getRegion()
        a, a_w = (x1 + x2) / 2, x2 - x1

        self.roi.angle = a
        self.roi.angle_std = a_w

        self.sigRoiMoved.emit(self.roi.key)



# -*- coding: utf-8 -*-
import logging

from PyQt5.QtWidgets import (QSlider, QLineEdit, QHBoxLayout,
                             QLabel, QWidgetAction, QMenu,
                             QSizePolicy, QWidget)
from PyQt5.QtGui import QPainter, QPainterPath, QColor, QIntValidator
from PyQt5.QtCore import Qt, pyqtSignal, QSize, QRectF

import numpy as np

from .buttons import RoundedPushButton

from ..tools import color_animation

logger = logging.getLogger(__name__)


class DoubleSlider(QSlider):
    valueChangedByHand = pyqtSignal(float)
    _MAX_INT = int(1e9)

    def __init__(self, orientation=Qt.Horizontal, parent=None, decimals: int = 5,
                 value: float = 0, bounds: tuple = (0, 1), log_scale: bool = False):
        super().__init__(orientation=orientation, parent=parent)
        self.decimals: int = decimals
        self._pressed: bool = False
        self._log_scale: bool = log_scale
        self.sliderPressed.connect(self._set_pressed)
        self.sliderReleased.connect(self._set_released)
        self.sliderReleased.connect(self.emit_value)
        self.valueChanged.connect(self._check_and_emit)

        super().setMinimum(0)
        super().setMaximum(self._MAX_INT)

        self._min_value = bounds[0]
        self._max_value = bounds[1]
        if self._log_scale and bounds[0] <= 0:
            raise ValueError(f'Wrong minimal bound for log scale: {bounds[0]}')
        self.setValue(value)

    def emit_value(self, *args):
        self.valueChangedByHand.emit(self.value())

    def _set_pressed(self, *args):
        self._pressed = True

    def _set_released(self, *args):
        self._pressed = False

    def _check_and_emit(self, *args):
        if self._pressed:
            self.emit_value()

    def set_decimals(self, decimals):
        self.decimals = decimals

    @property
    def _value_range(self):
        return self._max_value - self._min_value

    def _real_to_view(self, value):
        try:
            if self._log_scale:
                return int((np.log10(value) - np.log10(self._min_value)) /
                           (np.log10(self._max_value) - np.log10(self._min_value)) * self._MAX_INT)
            else:
                return int((value - self._min_value) / self._value_range * self._MAX_INT)
        except ZeroDivisionError:
            return 0

    def _view_to_real(self, value):
        if self._log_scale:
            return 10 ** (value / self._MAX_INT * (np.log10(self._max_value) - np.log10(self._min_value)) +
                          np.log10(self._min_value))
        else:
            return value / self._MAX_INT * self._value_range + self._min_value

    def value(self):
        return self._view_to_real(super().value())

    def setValue(self, value):
        super().setValue(self._real_to_view(value))

    def setMinimum(self, value):
        if value > self._max_value:
            raise ValueError("Minimum limit cannot be higher than maximum")
        if self._log_scale and value <= 0:
            raise ValueError(f'Wrong minimal bound for log scale: {value}')
        real_value = self.value()
        self._min_value = value
        self.setValue(real_value)
        if real_value < value:
            self.emit_value()

    def setMaximum(self, value):
        if value < self._min_value:
            raise ValueError("Minimum limit cannot be higher than maximum")

        real_value = self.value()
        self._max_value = value
        self.setValue(real_value)
        if real_value > value:
            self.emit_value()

    def setRange(self, p_int, p_int_1):
        real_value = self.value()
        self._min_value = p_int
        self._max_value = p_int_1
        if self._log_scale and p_int <= 0:
            raise ValueError(f'Wrong minimal bound for log scale: {p_int}')
        self.setValue(real_value)

    def minimum(self):
        return self._min_value

    def maximum(self):
        return self._max_value


class LabeledSlider(QWidget):
    valueChanged = pyqtSignal(float)

    _HEIGHT = 50
    _WIDTH_MIN = 200
    _WIDTH_MAX = 300

    def __init__(self, name: str, bounds: tuple = (0, 1),
                 value: float = 0,
                 parent=None, orientation=Qt.Horizontal, decimals: int = 3, scientific: bool = False,
                 log_scale: bool = False):
        super().__init__(parent=parent)
        self.name = name
        self._decimals = decimals
        self._scientific: bool = scientific
        self._bounds = bounds
        self.slider = DoubleSlider(orientation, parent, decimals, value, bounds, log_scale=log_scale)
        self.line_edit = QLineEdit(self.get_str_value())
        self.label = QLabel(name)

        self.line_edit.editingFinished.connect(self._set_value_from_text)
        self.line_edit.setStyleSheet('QLineEdit {  border: none; }')

        self.slider.valueChangedByHand.connect(self._set_value_from_slider)

        layout = QHBoxLayout(self)
        layout.addWidget(self.label)
        layout.addWidget(self.slider)
        layout.addWidget(self.line_edit)

        self.setSizePolicy(
            QSizePolicy.MinimumExpanding,
            QSizePolicy.Fixed
        )

        self.setFixedHeight(self._HEIGHT)
        self.edit_width = self.fontMetrics().width(
            ' '.join([' '] * (self._decimals + 4))) + int(self._scientific) * 50
        self.line_edit.setMaximumWidth(self.edit_width)
        self.setMaximumWidth(self._WIDTH_MAX)

    def sizeHint(self) -> QSize:
        return QSize(self._WIDTH_MIN, self._HEIGHT)

    @property
    def value(self) -> float:
        return self.slider.value()

    def get_str_value(self) -> str:
        if self._decimals:
            return f'{self.slider.value():.{self._decimals}{"e" if self._scientific else "f"}}'
        else:
            return str(int(self.slider.value()))

    def _set_value_from_slider(self):
        self.line_edit.setText(self.get_str_value())
        self.valueChanged.emit(self.slider.value())

    def _set_value_from_text(self):
        value = self.line_edit.text()
        try:
            if self._decimals:
                value = float(value)
            else:
                value = int(value)
            self.slider.setValue(value)
            if value < self._bounds[0] or value > self._bounds[1]:
                raise ValueError()
        except (ValueError, OverflowError):
            self.line_edit.setText(self.get_str_value())
            color_animation(self.line_edit)

        self.valueChanged.emit(self.slider.value())

    def set_value(self, value: float, change_bounds: bool = False):
        if change_bounds and value < self.slider.minimum():
            self.slider.setMinimum(value)

        elif change_bounds and value > self.slider.maximum():
            self.slider.setMaximum(value)

        self.slider.setValue(value)
        self.line_edit.setText(self.get_str_value())
        if value != self.value:
            self.valueChanged.emit(value)


class AnimatedSlider(RoundedPushButton):
    # TODO: change decimals to scientific notation!
    valueChanged = pyqtSignal(float)
    statusChanged = pyqtSignal(str)
    _EditMaximumWidth = 80
    _Height = 40
    _padding = 50

    @property
    def value(self):
        return self.slider.value()

    def __init__(self,
                 name: str,
                 bounds: tuple = (0, 1),
                 value: float = 0,
                 parent=None,
                 orientation=Qt.Horizontal,
                 decimals: int = 0,
                 hide: bool = True,
                 context_menu_enabled: bool = True,
                 disable_changing_status: bool = False,
                 min_max_bounds: tuple = (-10000, 10000)):

        super(AnimatedSlider, self).__init__(parent)
        self.context_menu_enabled = context_menu_enabled
        self.name = name
        self.bounds = bounds
        self._decimals = decimals
        self.init_value = value
        self.__init_ui__(orientation)
        self.clicked.connect(self.on_clicked)
        self._status = 'show'
        self._min_max_bounds = min_max_bounds
        self._disable_changing_status = disable_changing_status
        if hide:
            self.hide_and_show('hide')

    def set_min_max_bounds(self, bounds: tuple):
        if bounds[0] < bounds[1]:
            self._min_max_bounds = bounds

    def on_clicked(self):
        self.hide_and_show()
        self.statusChanged.emit(self._status)

    def contextMenuEvent(self, *args, **kwargs):
        if not self.context_menu_enabled:
            return
        event = args[0]
        menu = QMenu()
        menu.setWindowFlags(menu.windowFlags() | Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
        menu.setAttribute(Qt.WA_TranslucentBackground)
        menu.setStyleSheet("QMenu{background:rgba(255, 255, 255, 0%);}")

        decimals_slider = AnimatedSlider(
            'Decimals', (0, 6), self._decimals, hide=False,
            context_menu_enabled=False,
            disable_changing_status=True
        )
        min_slider = AnimatedSlider(
            'Min value', (self._min_max_bounds[0], self.slider.maximum()),
            self.slider.minimum(), hide=False,
            context_menu_enabled=False, disable_changing_status=True)
        max_slider = AnimatedSlider(
            'Max value', (self.slider.minimum(), self._min_max_bounds[1]),
            self.slider.maximum(), hide=False,
            context_menu_enabled=False, disable_changing_status=True)
        decimals_slider.valueChanged.connect(self.set_decimals)
        min_slider.valueChanged.connect(self.slider.setMinimum)
        min_slider.valueChanged.connect(max_slider.slider.setMinimum)
        max_slider.valueChanged.connect(self.slider.setMaximum)

        # TODO: _x - 1e-45 looks like the worst decision ever and should be well tested or better changed.
        max_slider.valueChanged.connect(
            lambda x: min_slider.slider.setMaximum(x - 1e-45))

        for w in [decimals_slider, min_slider, max_slider]:
            w_action = QWidgetAction(self)
            w_action.setDefaultWidget(w)
            menu.addAction(w_action)
        menu.exec_(self.mapToGlobal(event.pos()))

    def set_decimals(self, value):
        try:
            value = int(value)
        except ValueError as err:
            logger.exception(err)
            return
        self.slider.set_decimals(value)
        self._decimals = value
        self.editValue.setText(self._get_str_value())
        self._update_label_text()

    def set_bounds(self, bounds: tuple, change_value: bool = False):
        try:
            self.slider.setRange(*bounds)
        except ValueError as err:
            if not change_value:
                logger.exception(err)
            else:
                self.slider.setValue(bounds[0])
                self.slider.setRange(*bounds)

    def hide_and_show(self, new_status: str = None):
        if self._disable_changing_status:
            return
        if new_status in 'show hide'.split():
            self._status = new_status
        else:
            self._status = 'hide' if self._status == 'show' else 'show'

        if self._status == 'hide':
            self.slider.hide()
            self.editValue.hide()
            self._update_label_text()
        else:
            self.slider.show()
            self.editValue.show()
            self.label.setText(self.name)

    def _update_label_text(self, *args):
        if self._status == 'hide':
            new_text = f'{self.name}{" = " if self.name else ""}{self._get_str_value()}'
            self.label.setText(new_text)

    def _get_str_value(self):
        if self._decimals:
            return f'{self.slider.value():.{self._decimals}f}'
        else:
            return str(int(self.slider.value()))

    def set_value(self, value: float, change_bounds: bool = False):
        emit_value = False
        if value < self.slider.minimum():
            if change_bounds:
                self.slider.setMinimum(value)
            else:
                emit_value = True
        if value > self.slider.maximum():
            if change_bounds:
                self.slider.setMaximum(value)
            else:
                emit_value = True
        self.slider.setValue(value)
        self.editValue.setText(self._get_str_value())
        if emit_value:
            self.slider.emit_value()

    def __init_ui__(self, orientation):
        layout = QHBoxLayout(self)

        self.slider = DoubleSlider(orientation, self,
                                   decimals=self._decimals)

        self.slider.setRange(*self.bounds)
        self.slider.setValue(self.init_value)
        self.slider.valueChangedByHand.connect(
            lambda *x: self.valueChanged.emit(self.slider.value()))
        self.editValue = QLineEdit(self._get_str_value())
        self.editValue.setMaximumWidth(self._EditMaximumWidth)
        self.editValue.setStyleSheet('QLineEdit {  border: none; }')
        self.label = QLabel(self.name)
        self.slider.valueChanged.connect(self._update_label_text)

        if not self._decimals:
            self.editValue.setValidator(QIntValidator(*self.bounds))

        self.slider.valueChangedByHand.connect(
            lambda _: self.editValue.setText(self._get_str_value()))
        self.editValue.editingFinished.connect(self._set_slider_value)

        layout.addWidget(self.label)
        layout.addWidget(self.slider)
        layout.addWidget(self.editValue)

        self.setFixedHeight(self._Height)
        min_width = self.fontMetrics().width(
            ' = '.join([self.name, self._get_str_value()])) + self._padding
        self.setMinimumWidth(min_width)

    def _set_slider_value(self):
        value = self.editValue.text()
        try:
            if self._decimals:
                value = float(value)
            else:
                value = int(value)
            self.slider.setValue(value)
            if value < self.bounds[0] or value > self.bounds[1]:
                raise ValueError()
        except ValueError:
            self.editValue.setText(self._get_str_value())
            color_animation(self.editValue)
        self.valueChanged.emit(self.slider.value())


class ControlSlider(AnimatedSlider):
    def __init__(self, *args, **kwargs):
        AnimatedSlider.__init__(self, *args, **kwargs)
        self._fixed = False

    def set_fixed(self):
        self.hide_and_show('hide')
        self._fixed = True
        self.context_menu_enabled = False
        self.label.setStyleSheet('QLabel {color: green; }')

    def set_unfixed(self):
        self._fixed = False
        self.context_menu_enabled = True
        self.label.setStyleSheet('QLabel {color: black; }')

    def on_clicked(self):
        if not self._fixed:
            AnimatedSlider.on_clicked(self)


class ParametersSlider(QWidget):
    sigLowerValueChanged = pyqtSignal(float)
    sigMiddleValueChanged = pyqtSignal(float)
    sigUpperValueChanged = pyqtSignal(float)

    _rect_width = 10
    _rect_height = 20
    _padding = 30
    _slider_width = 5

    # _path_color = QColor('black')
    _rect_color = QColor(100, 100, 255)

    def __init__(self, x1=0., x2=0.5, x3=0.1,
                 min_value=0, max_value=1, parent=None):
        super().__init__(parent=parent)

        # self.setMouseTracking(True)
        self.setSizePolicy(
            QSizePolicy.Expanding,
            QSizePolicy.Minimum
        )

        self._pressed: int = 0

        self._min_value = min_value
        self._max_value = max_value

        self._x1 = x1
        self._x2 = x2
        self._x3 = x3

    def setValues(self, x1, x2, x3, *, adjust_range: bool = False, new_range: tuple = None):
        if x1 <= x2 <= x3:
            if adjust_range:
                self._min_value = min(self._min_value, x1)
                self._max_value = max(self._max_value, x3)
            elif new_range:
                self._min_value, self._max_value = new_range
                if self._min_value > self._max_value:
                    raise ValueError(f'Wrong range: {self._min_value} > {self._max_value}')

            if x1 < self._min_value or x3 > self._max_value:
                raise ValueError('Out of range.')

            self._x1 = x1
            self._x2 = x2
            self._x3 = x3

            self.update()

        else:
            raise ValueError(f'Wrong order: {x1}, {x2}, {x3}')

    def sizeHint(self):
        return QSize(300, self._rect_height)

    def mousePressEvent(self, ev):
        if ev.button() == Qt.RightButton:
            ev.ignore()
        else:
            x = ev.pos().x()
            self._pressed = self._get_idx(x)

            if self._pressed:
                ev.accept()
            else:
                ev.ignore()

    def mouseReleaseEvent(self, ev):
        if self._pressed:
            ev.accept()
            self._pressed = 0
        else:
            ev.ignore()

    def mouseMoveEvent(self, ev):
        if not self._pressed:
            # idx = self._get_idx(ev.pos().x())
            # if idx:
            #     print('Hover: ', idx)
            #     ev.accept()
            # else:
            ev.ignore()

        else:
            x = ev.pos().x()
            bounds = self._get_bounds(self._pressed)

            if x < bounds[0]:
                self._set_value(bounds[0])
            elif x > bounds[1]:
                self._set_value(bounds[1])
            else:
                self._set_value(x)

    def _get_idx(self, x):
        w = self._rect_width / 2

        if abs(self._x1_view - x) <= w:
            return 1
        elif abs(self._x2_view - x) <= w:
            return 2
        elif abs(self._x3_view - x) <= w:
            return 3
        else:
            return 0

    def _get_bounds(self, idx: int):
        if not idx:
            return
        w = self._rect_width

        if idx == 1:
            return self._padding, self._x2_view - w
        elif idx == 2:
            return self._x1_view + w, self._x3_view - w
        else:
            return self._x2_view + w, self.width() - self._padding

    def _set_value(self, x):
        idx = self._pressed
        if not idx:
            return

        value = self._scale_from_view(x, idx)

        if idx == 1 and value != self._x1:
            self._x1 = value
            self.sigLowerValueChanged.emit(self._x1)
            self.update()
        elif idx == 2 and value != self._x2:
            self._x2 = value
            self.sigMiddleValueChanged.emit(self._x2)
            self.update()
        elif idx == 3 and value != self._x3:
            self._x3 = value
            self.sigUpperValueChanged.emit(self._x3)
            self.update()

    def paintEvent(self, ev):

        p = QPainter(self)

        self._x1_view = self._scale_to_view(self._x1, 1)
        self._x2_view = self._scale_to_view(self._x2, 2)
        self._x3_view = self._scale_to_view(self._x3, 3)

        p.setRenderHint(QPainter.Antialiasing)

        self._draw_slider(p)
        self._draw_rect(self._x1_view, p)
        self._draw_rect(self._x2_view, p)
        self._draw_rect(self._x3_view, p)

        p.end()

    def _draw_slider(self, p: QPainter):
        path = QPainterPath()
        path.addRoundedRect(
            QRectF(self._padding, self.height() / 2 - self._slider_width / 2,
                   self.width() - self._padding * 2, self._slider_width),
            3, 3)
        # pen = QPen(self._path_color, 2)
        # p.setPen(pen)
        p.fillPath(path, self._rect_color)
        p.drawPath(path)

    def _draw_rect(self, x: float, p: QPainter):
        path = QPainterPath()
        path.addRoundedRect(
            QRectF(x - self._rect_width / 2,
                   self.height() / 2 - self._rect_height / 2,
                   self._rect_width, self._rect_height),
            1, 1)
        # pen = QPen(self._path_color, 2)
        # p.setPen(pen)
        p.fillPath(path, self._rect_color)
        p.drawPath(path)

    @property
    def _length(self):
        return self.width() - 2 * self._padding - 2 * self._rect_width

    @property
    def _range(self):
        return self._max_value - self._min_value

    def _scale_to_view(self, x, idx: int):
        return self._padding + self._rect_width * (idx - 1) + \
               (x - self._min_value) / self._range * self._length

    def _scale_from_view(self, x, idx: int):
        return self._min_value + (x - self._padding - self._rect_width * (idx - 1)) * \
               self._range / self._length



import logging
import pickle
from pathlib import Path

from h5py import Group, File

from ..utils import InternalError
from .keys import ImageKey, AbstractKey, ImageH5Key
from .project_structure import ProjectStructure


def _check_empty_project(func):
    def wrapper(self, *args, **kwargs):
        if self.project_structure.project_opened:
            return func(*args, **kwargs)
        else:
            return

    return wrapper


class _ObjectFileManager(object):
    log = logging.getLogger(__name__)

    NAME = ''

    def __init__(self, project_structure: ProjectStructure):
        self.project_structure = project_structure
        self.folder: Path = None
        self.init()

    def init(self):
        path = self.project_structure.path

        if path:
            self.folder: Path = path / self.NAME
            self.folder.mkdir(parents=True, exist_ok=True)

    def _get_path(self, key: AbstractKey) -> Path:
        return self.folder / key.file_name()

    @staticmethod
    def _set_pickle(path: Path, value):
        with open(str(path.resolve()), 'wb') as f:
            pickle.dump(value, f)

    @staticmethod
    def _get_pickle(path: Path):
        if path.is_file():
            with open(str(path.resolve()), 'rb') as f:
                return pickle.load(f)

    @staticmethod
    def _del_pickle(path: Path):
        if path.is_file():
            path.unlink()

    @staticmethod
    def _get_h5(h5group: Group, key: ImageKey):
        pass

    @staticmethod
    def _set_h5(h5group: Group, key: ImageKey, *args, **kwargs):
        pass

    @staticmethod
    def _del_h5(h5group: Group, key: ImageKey):
        pass

    def _process_h5_group(self, h5path: Path, h5key: str, func, *args, **kwargs):
        try:
            with File(str(h5path.resolve()), 'r') as f:
                return func(f[h5key], *args, **kwargs)
        except FileNotFoundError:
            raise FileNotFoundError(f'H5 file {h5path} is not found!')
        except (KeyError, IOError):
            raise IOError(f'Project H5 file {h5path} is corrupted!')
        except Exception as err:
            self.log.exception(err)
            raise InternalError(f'An error occurred while reading project'
                                f'h5 file {h5path}.')

    # def _process_pickle_file(self, path: Path, func, *args, **kwargs):
    #     if not path.is_file():
    #         return
    #     try:
    #         return func(path, *args, **kwargs)
    #     except Exception as err:
    #         self.log.exception(err)
    #         raise InternalError(f'An error occurred while reading project'
    #                             f' internal files.')

    def __getitem__(self, key):
        if not key.is_project:
            return self._get_pickle(self._get_path(key))
        if self.project_structure.config[key.h5path]:
            return self._process_h5_group(key.h5path, key.h5key, self._get_h5, key)
        res = self._get_pickle(self._get_path(key))
        if res is not None:
            return res
        return self._process_h5_group(key.h5path, key.h5key, self._get_h5, key)

    def __delitem__(self, key):
        if not key.is_project:
            return self._del_pickle(self._get_path(key))
        if self.project_structure.config[key.h5path]:
            return self._process_h5_group(key.h5path, key.h5key, self._del_h5, key)
        else:
            return self._del_pickle(self._get_path(key))

    def __setitem__(self, key, value):
        if key.is_project and self.project_structure.config[key.h5path]:
            return self._process_h5_group(key.h5path, key.h5key, self._set_h5, key, value)
        else:
            try:
                return self._set_pickle(self._get_path(key), value)
            except Exception as err:
                self.log.exception(err)
                return
            # return self._process_pickle_file(self._get_path(key), self._set_pickle, value)

    def _check_save_to_h5(self, key) -> bool:
        if key.is_project and self.project_structure.config[key.h5path]:
            return True
        return False



from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot

from pyqtgraph import RectROI

from ...app import Roi, App
from .abstract_roi_widget import AbstractRoiWidget


class Roi2DRect(AbstractRoiWidget, RectROI):
    sigRoiMoved = pyqtSignal(int)
    sigSelected = pyqtSignal(int)
    sigShiftSelected = pyqtSignal(int)

    def __init__(self, roi: Roi, enable_context: bool = True, **kwargs):
        AbstractRoiWidget.__init__(self, roi, enable_context=enable_context, **kwargs)
        RectROI.__init__(self, pos=(0, 0), size=(1, 1), centered=False, sideScalers=False)

        self.addScaleHandle([0.5, 0], [0.5, 1])
        self.addScaleHandle([0.5, 1], [0.5, 0])
        self.addScaleHandle([0, 0.5], [1, 0.5])
        self.addScaleHandle([1, 0.5], [0, 0.5])

        self.handle = self.handles[0]['item']
        self.handles.pop(0)
        self.handle.disconnectROI(self)
        self.handle.hide()  # how to remove???
        self.sigRegionChanged.connect(self._handle_is_moving)
        self.update_roi()

        if App().debug_tracker:
            App().debug_tracker.add_object(self, roi.name)


    def _handle_is_moving(self):
        for h in self.handles:
            if h['item'].isMoving:
                self.send_move()
                return

    def _hide_handles(self):
        for h in self.handles:
            h['item'].hide()

    def _show_handles(self):
        for h in self.handles:
            h['item'].show()

    def send_move(self):
        size, pos = self.size(), self.pos()
        w, a_w = size
        r, a = abs(pos[0] + w / 2), pos[1] + a_w / 2
        self.roi.radius = r
        self.roi.width = w
        self.roi.angle = a
        self.roi.angle_std = a_w
        self.sigRoiMoved.emit(self.roi.key)

    @pyqtSlot(name='move_roi')
    def move_roi(self):
        r, w = self.roi.radius, abs(self.roi.width)
        a, a_w = self.roi.angle, abs(self.roi.angle_std)

        pos = [r - w / 2, a - a_w / 2]
        size = [w, a_w]
        self.setSize(size)
        self.setPos(pos)

    def set_color(self, color):
        self.setPen(color)

    def set_movable(self, movable: bool):
        self.translatable = movable

    def mouseDragEvent(self, ev):
        if self.roi.movable:
            RectROI.mouseDragEvent(self, ev)
            self.send_move()

    def mouseClickEvent(self, ev):
        if ev.button() == Qt.RightButton:
            ev.accept()
            self.show_context_menu(ev)
        elif ev.button() == Qt.LeftButton and ev.modifiers() == Qt.ShiftModifier:
            ev.accept()
            self.sigShiftSelected.emit(self.roi.key)
        elif ev.button() == Qt.LeftButton:
            # self.change_active()
            ev.accept()
            self.sigSelected.emit(self.roi.key)

    def fix(self):
        super().fix()
        self.set_movable(False)
        self._hide_handles()

    def unfix(self):
        super().unfix()
        self.set_movable(True)
        self._show_handles()



from PyQt5.QtGui import (QPainter, QPixmap, QPen, QBrush,
                         QConicalGradient, QColor)
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import (Qt, QPropertyAnimation, pyqtProperty,
                          pyqtSlot, QRect)


class Spinner(QWidget):
    def __init__(self, parent=None, size: int = 25, padding: int = 2, start: bool = True):
        super().__init__(parent)
        self._size = size
        self._padding = padding

        self.pixmap = QPixmap('wheel.png').scaled(size, size)

        self.setFixedSize(size + padding * 2, size + padding * 2)
        self.setWindowFlags(Qt.FramelessWindowHint)

        self._angle = 0

        self.animation = QPropertyAnimation(self, b"angle", self)
        self.animation.setStartValue(0)
        self.animation.setEndValue(360)
        self.animation.setLoopCount(-1)
        self.animation.setDuration(2000)
        if start:
            self.animation.start()

    @pyqtSlot(name='pauseSpinning')
    def pause(self):
        self.animation.stop()

    @pyqtSlot(name='resumeSpinning')
    def resume(self):
        self.animation.start()

    @pyqtProperty(int)
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value
        self.update()

    def paintEvent(self, ev=None):
        drawing_rect = QRect()
        drawing_rect.setX(self._padding)
        drawing_rect.setY(self._padding)
        drawing_rect.setWidth(self._size)
        drawing_rect.setHeight(self._size)

        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        gradient = QConicalGradient()
        gradient.setCenter(drawing_rect.center())
        gradient.setAngle(- self._angle - self._size / 10)
        gradient.setColorAt(0, QColor(178, 255, 246))
        gradient.setColorAt(1, QColor(5, 44, 50))

        pen = QPen(QBrush(gradient), self._size // 10)
        pen.setCapStyle(Qt.RoundCap)
        painter.setPen(pen)
        painter.drawArc(drawing_rect, -self._angle * 16, 300 * 16)



from typing import Tuple, NamedTuple
from copy import deepcopy

import numpy as np

from .transformations import TransformationsHolder


class AxesScale(object):
    def __init__(self, scale: float = 1):
        self._scale = scale
        self._prev_scale = 1.

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        self.set_scale(value)

    @property
    def scale_change(self):
        return self.scale / self._prev_scale

    def set_scale(self, scale: float):
        self._scale, self._prev_scale = scale, self._scale


class BeamCenter(NamedTuple):
    z: float = 0
    y: float = 0


class Geometry(object):
    def __init__(self, *, beam_center: tuple = (0, 0),
                 scale: float = 1.,
                 shape: Tuple[int, int] = (10, 10),
                 polar_shape: Tuple[int, int] = (512, 512),
                 t_key: str = None, update: bool = True,
                 **kwargs):

        self._beam_center = BeamCenter(*beam_center)
        self._scale = AxesScale(scale)
        self._transforms = TransformationsHolder()
        if t_key:
            self._transforms.update(t_key)
        self._shape = tuple(shape)
        self._polar_shape = tuple(polar_shape)

        self._r_range = (0, 1)
        self._phi_range = (0, 2 * np.pi)
        self._ring_bounds = (0, 2 * np.pi)
        self._r = self._phi = None
        self._y = self._z = None
        self._polar_zz = self._polar_yy = None
        self._polar_aspect_ratio = None

        if update:
            self.update()

    def to_dict(self) -> dict:
        return dict(beam_center=tuple(self.beam_center),
                    shape=self.shape,
                    scale=self.scale,
                    t_key=self.t.key,
                    polar_shape=self.polar_shape)

    @staticmethod
    def keys():
        return 'beam_center', 'shape', 'scale', 't_key', 'polar_shape'

    @classmethod
    def fromdict(cls, d: dict):
        return cls(**d)

    def __eq__(self, other):
        if type(other) != Geometry:
            return False

        return (self.beam_center == other.beam_center and
                self.scale == other.scale and
                self.shape == other.shape and
                self.t.key == other.t.key)

    @property
    def t(self):
        return self._transforms

    def copy(self) -> 'Geometry':
        return deepcopy(self)

    @property
    def is_available(self) -> bool:
        return self.shape is not None

    @property
    def ring_bounds(self):
        return self._ring_bounds

    @property
    def beam_center(self):
        return self._beam_center

    @property
    def r_range(self) -> Tuple[float, float]:
        return self._r_range

    @property
    def phi_range(self) -> Tuple[float, float]:
        return self._phi_range

    @property
    def polar_grids(self):
        return self._polar_yy, self._polar_zz

    @property
    def r_axis(self):
        return self._r

    @property
    def phi_axis(self):
        return self._phi

    @property
    def polar_aspect_ratio(self):
        return self._polar_aspect_ratio

    @property
    def y_axis(self):
        return self._y

    @property
    def z_axis(self):
        return self._z

    @property
    def shape(self):
        return self._shape

    @property
    def polar_shape(self):
        return self._polar_shape

    @property
    def scale(self):
        return self._scale.scale

    @property
    def scale_change(self):
        return self._scale.scale_change

    def set_scale(self, scale: float, update: bool = True):
        if scale != self.scale:
            self._scale.set_scale(scale)
            if update:
                self._update_axes_on_scale(False)

    def set_beam_center(self, z: float, y: float, update: bool = True):
        if (self._beam_center.z, self._beam_center.y) != (z, y):
            self._beam_center = BeamCenter(z, y)
            if update:
                self.update()
            return True
        return False

    def set_shape(self, shape: Tuple[int, int], update: bool = True):
        if shape != self.shape:
            self._shape = shape
            if update:
                self.update()
            return True
        return False

    def set_polar_shape(self, shape: Tuple[int, int], update: bool = True):
        if shape != self.polar_shape:
            self._polar_shape = shape
            if update:
                self.update_polar()
            return True
        return False

    def update(self):
        if not self.is_available:
            return
        self._update_ranges()
        self._update_polar_grid()
        self._update_axes_on_scale()

    def update_polar(self):
        self._update_polar_grid()
        self._r *= self.scale
        self._polar_aspect_ratio /= self.scale

    def _update_ranges(self):
        self._y = (np.arange(self._shape[1]) - self._beam_center.y)
        self._z = (np.arange(self._shape[0]) - self._beam_center.z)

        yy, zz = np.meshgrid(self._y, self._z)
        rr = np.sqrt(yy ** 2 + zz ** 2)
        phi = np.arctan2(zz, yy)
        self._r_range = (rr.min(), rr.max())
        self._phi_range = p_min, p_max = phi.min(), phi.max()
        angle, angle_std = (p_max + p_min) / 2 * 180 / np.pi, (p_max - p_min) * 180 / np.pi
        self._ring_bounds = (angle, angle_std)

    def _update_polar_grid(self):
        self._phi = np.linspace(*self.phi_range, self.polar_shape[0])
        self._r = np.linspace(*self.r_range, self.polar_shape[1])

        r_matrix = self._r[np.newaxis, :].repeat(self.polar_shape[0], axis=0)
        p_matrix = self._phi[:, np.newaxis].repeat(self.polar_shape[1], axis=1)

        self._polar_yy = r_matrix * np.cos(p_matrix) + self.beam_center.y
        self._polar_zz = r_matrix * np.sin(p_matrix) + self.beam_center.z

        self._phi *= 180 / np.pi

        aspect_ratio = (self._phi.max() - self._phi.min()) * \
                       self.polar_shape[0] / (self._r_range[1] - self._r_range[0]) / self.polar_shape[1]
        self._polar_aspect_ratio = aspect_ratio

    def _update_axes_on_scale(self, init: bool = True):
        if init:
            scale = self.scale
        else:
            scale = self.scale_change

        self._r = self._r * scale
        self._y = self._y * scale
        self._z = self._z * scale
        self._polar_aspect_ratio = self._polar_aspect_ratio / scale



from h5py import Group

from .object_file_manager import _ObjectFileManager
from ..geometry import Geometry


class _DefaultGeometry(_ObjectFileManager):
    NAME = 'geometries'

    def _get_path(self, key):
        return self.folder / f'default_geometry_{key.parent.name}'

    @staticmethod
    def _get_h5(h5group: Group, key):
        return Geometry.fromdict(dict(h5group.parent.attrs))

    @staticmethod
    def _set_h5(h5group: Group, key, value: Geometry):
        h5group.parent.attrs.update(value.to_dict())

    @staticmethod
    def _del_h5(h5group: Group, key):
        for key in Geometry.keys():
            if key in h5group.parent.attrs.keys():
                del h5group.attrs[key]


class _ReadGeometry(_ObjectFileManager):
    NAME = 'geometries'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = _DefaultGeometry(*args, **kwargs)

    @staticmethod
    def _get_h5(h5group: Group, key):
        return Geometry.fromdict(dict(h5group.attrs))

    @staticmethod
    def _set_h5(h5group: Group, key, value: Geometry):
        h5group.attrs.update(value.to_dict())

    @staticmethod
    def _del_h5(h5group: Group, key):
        for key in Geometry.keys():
            if key in h5group.attrs.keys():
                del h5group.attrs[key]



# -*- coding: utf-8 -*-

from PyQt5.QtGui import QCursor
from PyQt5.QtWidgets import QMenu, QWidgetAction, QLineEdit

from ...app.app import App
from ...app.rois.roi import Roi, RoiTypes


class AbstractRoiContextMenu(QMenu):

    def __init__(self, roi: Roi):
        QMenu.__init__(self)
        self.roi = roi
        self.key = roi.key
        self.roi_dict = App().roi_dict
        self._init_menu()

        if App().debug_tracker:
            App().debug_tracker.add_object(self)

        self.exec_(QCursor.pos())
        # self.exec_(ev.globalPos())


class RoiContextMenu(AbstractRoiContextMenu):
    def _init_menu(self):
        self._init_rename_menu()
        self.addSeparator()
        self._init_type_menu()
        self.addSeparator()
        self._init_fix_menu()
        self.addSeparator()
        self._init_copy_menu()
        self.addSeparator()
        self._init_select_menu()
        self.addSeparator()
        self._init_fit_menu()
        self.addSeparator()
        self._init_delete_menu()

    def _init_fix_menu(self):
        fix_menu = self.addMenu('Fix/Unfix')
        if self.roi.movable:
            fix_action = fix_menu.addAction('Fix roi')
            fix_action.triggered.connect(lambda: self.roi_dict.fix_roi(self.key))
        else:
            fix_action = fix_menu.addAction('Unfix roi')
            fix_action.triggered.connect(lambda: self.roi_dict.unfix_roi(self.key))

        fix_selected = fix_menu.addAction('Fix selected roi')
        fix_selected.triggered.connect(self.roi_dict.fix_selected)
        unfix_selected = fix_menu.addAction('Unfix selected roi')
        unfix_selected.triggered.connect(self.roi_dict.unfix_selected)

        fix_all = fix_menu.addAction('Fix all roi')
        fix_all.triggered.connect(self.roi_dict.fix_all)

        unfix_all = fix_menu.addAction('Unix all roi')
        unfix_all.triggered.connect(self.roi_dict.unfix_all)

    def _init_delete_menu(self):
        delete_menu = self.addMenu('Delete')
        delete_self = delete_menu.addAction('Delete this roi')
        delete_self.triggered.connect(lambda: self.roi_dict.delete_roi(self.key))
        delete_selected = delete_menu.addAction('Delete selected')
        delete_selected.triggered.connect(self.roi_dict.delete_selected_roi)

    def _init_select_menu(self):
        select_menu = self.addMenu('Select')
        select_all = select_menu.addAction('Select all')
        select_all.triggered.connect(self.roi_dict.select_all)
        unselect_all = select_menu.addAction('Unselect all')
        unselect_all.triggered.connect(self.roi_dict.unselect_all)

    def _init_rename_menu(self):
        rename = self.addMenu('Rename')
        rename_action = QWidgetAction(self)
        line_edit = QLineEdit(self.roi.name)
        line_edit.editingFinished.connect(
            lambda: self.roi_dict.change_name(self.key, line_edit.text())
        )
        rename_action.setDefaultWidget(line_edit)
        rename.addAction(rename_action)

    def _init_type_menu(self):
        if self.roi.type == RoiTypes.ring:
            new_type = RoiTypes.segment
            change_type_name = 'segment'
        else:
            new_type = RoiTypes.ring
            change_type_name = 'ring'
        change_type_action = self.addAction(f'Change type to {change_type_name}')
        change_type_action.triggered.connect(
            lambda: self._change_roi_type(new_type))

    def _change_roi_type(self, new_type: RoiTypes):
        self.roi.type = new_type
        self.roi_dict.change_roi_type(self.key)

    def _init_fit_menu(self):
        fit_menu = self.addMenu('Fit')
        fit_selected = fit_menu.addAction('Fit selected rois')
        fit_selected.triggered.connect(lambda: self.roi_dict.open_fit_rois(True))
        fit_selected = fit_menu.addAction('Fit all rois')
        fit_selected.triggered.connect(lambda: self.roi_dict.open_fit_rois(False))

    def _init_copy_menu(self):
        copy_menu = self.addMenu('Copy/Paste')
        copy_menu.addAction('Copy roi', lambda: self.roi_dict.copy_rois(self.roi.key))
        copy_menu.addAction('Copy selected rois', lambda: self.roi_dict.copy_rois('selected'))
        copy_menu.addAction('Copy all rois', lambda: self.roi_dict.copy_rois('all'))
        if self.roi_dict.is_copied:
            copy_menu.addAction('Paste rois', lambda: self.roi_dict.paste_rois())



import logging
import sys
import traceback

from PyQt5.QtCore import QObject, pyqtSlot, pyqtSignal, QRunnable

from scipy import sparse
from scipy.sparse.linalg import spsolve
from scipy.ndimage import gaussian_filter1d

import numpy as np

logger = logging.getLogger(__name__)


class SingletonMeta(type):
    _instance = None

    def __call__(cls):
        if cls._instance is None:
            cls._instance = super().__call__()
        return cls._instance


class InternalError(Exception):
    pass


def baseline_correction(y: np.ndarray,
                        smoothness_param: float,
                        asymmetry_param: float,
                        max_niter: int = 1000) -> np.ndarray:
    z = np.zeros_like(y)
    if smoothness_param <= 0 or asymmetry_param <= 0:
        return z
    y_size = y.size
    laplacian = sparse.diags([1, -2, 1], [0, -1, -2], shape=(y_size, y_size - 2))
    laplacian_matrix = laplacian.dot(laplacian.transpose())

    w = np.ones(y_size)
    for i in range(max_niter):
        W = sparse.spdiags(w, 0, y_size, y_size)
        Z = W + smoothness_param * laplacian_matrix
        z = spsolve(Z, w * y)
        w_new = asymmetry_param * (y > z) + (1 - asymmetry_param) * (y < z)
        if np.allclose(w, w_new):
            break
        w = w_new
    else:
        logger.info(f'Solution has not converged, max number of iterations reached.')
    return np.nan_to_num(z)


def smooth_curve(y: np.ndarray, sigma: float) -> np.ndarray or None:
    if y is not None:
        if sigma > 0:
            return gaussian_filter1d(y, sigma)
        else:
            return y


class WorkerSignals(QObject):
    finished = pyqtSignal()
    error = pyqtSignal(tuple)
    result = pyqtSignal(object)


class Worker(QRunnable):
    log = logging.getLogger(__name__)

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()

        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

    @pyqtSlot()
    def run(self):

        try:
            result = self.fn(
                *self.args, **self.kwargs
            )
        except Exception as err:
            self.log.exception(err)
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        else:
            self.signals.result.emit(result)
        finally:
            self.signals.finished.emit()



# -*- coding: utf-8 -*-

from dataclasses import dataclass
from typing import Tuple

import numpy as np
from scipy.optimize import curve_fit

from .functions import FittingType, FittingFunction, FITTING_FUNCTIONS
from .background import BackgroundType, BACKGROUNDS, Background
from .utils import Roi
from .range_strategy import RangeStrategyType, RangeStrategy


@dataclass
class Fit:
    roi: Roi

    r_range: Tuple[float, float]
    x_range: Tuple[int, int]
    x: np.ndarray
    y: np.ndarray

    init_curve: np.ndarray
    init_params: list
    lower_bounds: list
    upper_bounds: list

    fitted_params: list
    fit_errors: list
    fitting_curve: np.ndarray
    background_curve: np.ndarray

    fitting_function: FittingFunction
    background: Background
    range_strategy: RangeStrategy

    @property
    def param_names(self):
        return [*self.fitting_function.PARAM_NAMES, *self.background.PARAM_NAMES]

    @property
    def bounds(self):
        return tuple(self.lower_bounds), tuple(self.upper_bounds)

    def update(self, **params):
        for k, v in params.items():
            setattr(self, k, v)

    def update_fit(self, update_bounds: bool = True):
        if self.roi.movable:
            if update_bounds and self.x.size:
                self.init_params, self.upper_bounds, self.lower_bounds = self.fitting_function.bounds(
                    self.x, self.y, self.roi, self.background)
            else:
                self.fitting_function.set_roi_from_params(self.roi, self.init_params)

        self.background_curve = self.background(self.x, *self.init_params)
        self.init_curve = self.fitting_function(self.x, self.background, *self.init_params)

        if self.fitted_params:
            self.fitting_curve = self.fitting_function(self.x, self.background, *self.fitted_params)

    def set_background(self, background: Background, update: bool = True):
        self.background = background
        if update:
            self.update_fit()

    def set_function(self, fitting_function: FittingFunction, update: bool = True):
        self.fitting_function = fitting_function
        if update:
            self.update_fit()

    def do_fit(self) -> None:
        if not self.y.size or not self.x.size:
            return
        try:
            func = self.fitting_function.get_func(self.background)
            popt, pcov = curve_fit(func, self.x, self.y, self.init_params,
                                   bounds=self.bounds)
            perr = np.sqrt(np.diag(pcov))

            self.fitted_params = popt.tolist()
            self.init_params = self.fitted_params
            self.fit_errors = perr.tolist()
            self.fitting_curve = func(self.x, *popt)
            self.background_curve = self.background(self.x, *popt)
            self.init_curve = self.fitting_curve
            self.roi.fitted_parameters = dict(zip(self.param_names, self.fitted_params))
            self.roi.fitted_parameters['fitting_function'] = self.fitting_function.TYPE
            self.roi.fitted_parameters['background'] = self.background.TYPE
            self.fitting_function.set_roi_from_params(self.roi, self.fitted_params)

        except (ValueError, RuntimeError):
            return



# -*- coding: utf-8 -*-

from PyQt5.QtWidgets import QMainWindow, QLabel, QComboBox, QHBoxLayout
from PyQt5.QtCore import pyqtSlot, Qt

import numpy as np
from .basic_widgets import (CustomImageViewer, BlackToolBar, BasicInputParametersWidget,
                            InfoButton, AbstractInputParametersWidget)
from .roi_widgets.roi_2d_rect_widget import Roi2DRect
from .roi_widgets.abstract_roi_holder import AbstractRoiHolder

from ..app.app import App
from ..app.polar_image import INTERPOLATION_ALGORITHMS
from .tools import Icon


class PolarImageViewer(AbstractRoiHolder, QMainWindow):

    def __init__(self, parent=None):
        AbstractRoiHolder.__init__(self, 'PolarImageViewer')
        QMainWindow.__init__(self, parent)
        self.app = App()
        self._setup_window = None
        self._interpolation_params_dict: dict = InterpolateSetupWindow.get_config()

        self._image_viewer = CustomImageViewer(self)
        self._image_viewer.image_plot.getAxis('bottom').setLabel(
            text='|Q|', color='white', font_size='large')
        self._image_viewer.image_plot.getAxis('left').setLabel(
            text='&Phi;', color='white', font_size='large')
        self.setCentralWidget(self._image_viewer)
        self.app.image_holder.sigPolarImageChanged.connect(self._update_image)
        self.app.geometry_holder.sigScaleChanged.connect(self._on_scale_changed)
        self.app.image_holder.sigEmptyImage.connect(self._image_viewer.clear_image)
        # self.app.roi_dict.sig_roi_moved.connect(self._image_viewer.set_auto_range)

        self.__init_toolbar()

    def _on_scale_changed(self):
        self.set_axes()

    def _make_roi_widget(self, roi):
        roi_widget = Roi2DRect(roi)
        self._image_viewer.image_plot.addItem(roi_widget)
        return roi_widget

    def _delete_roi_widget(self, roi_widget):
        self._image_viewer.image_plot.removeItem(roi_widget)
        self._image_viewer.set_auto_range()

    def __init_toolbar(self):
        setup_toolbar = BlackToolBar('Setup', self)
        self.addToolBar(setup_toolbar)

        setup_action = setup_toolbar.addAction(Icon('setup'), 'Setup')
        setup_action.triggered.connect(self.open_setup_window)

    @pyqtSlot(name='updateImage')
    def _update_image(self):
        img = self.app.polar_image
        if img is not None:
            self._image_viewer.set_data(img)
            self.set_axes()

    def set_axes(self):
        if self._image_viewer.image_item.image is None:
            return
        p1, p2 = self.app.geometry.phi_range
        r1, r2 = self.app.geometry.r_axis.min(), self.app.geometry.r_axis.max()
        self._image_viewer.set_x_axis(r1, r2)
        self._image_viewer.set_y_axis(p1 * 180 / np.pi, p2 * 180 / np.pi)
        self._image_viewer.view_box.setAspectLocked(
            True, self.app.geometry.polar_aspect_ratio)
        self._image_viewer.set_auto_range()

    def open_setup_window(self):
        self._setup_window = InterpolateSetupWindow(params_dict=self.app.image_holder.polar_image_params_dict())
        self._setup_window.apply_signal.connect(self.set_fit_parameters)
        self._setup_window.close_signal.connect(self.close_peaks_setup)
        self._setup_window.show()

    def set_fit_parameters(self, params: dict):
        self._interpolation_params_dict = params
        self.app.image_holder.set_polar_image_params(params)

    def close_peaks_setup(self):
        self._setup_window = None


class InterpolateSetupWindow(BasicInputParametersWidget):
    P = BasicInputParametersWidget.InputParameters

    # TODO: add info to interpolation parameters

    PARAMETER_TYPES = (P('r_size', 'Radius axis size', int),
                       P('phi_size', 'Angle axis size', int),
                       P('mode', 'Interpolation algorithm', str))

    NAME = 'Interpolation parameters'

    DEFAULT_DICT = dict(r_size=512, phi_size=512, mode='Bilinear')

    def _get_layout(self,
                    input_parameter: BasicInputParametersWidget.InputParameters):
        if input_parameter.name == 'mode':
            return self._get_mode_layout(input_parameter)
        else:
            return super()._get_layout(input_parameter)

    def _get_mode_layout(self, input_parameter: BasicInputParametersWidget.InputParameters):

        current_value = self.default_dict.get(input_parameter.name, None)
        if current_value is None:
            current_value = next(iter(INTERPOLATION_ALGORITHMS.keys()))
        label_widget = QLabel(input_parameter.label)
        input_widget = QComboBox()
        input_widget.setEditable(False)
        input_widget.addItems(list(INTERPOLATION_ALGORITHMS.keys()))
        input_widget.setCurrentText(current_value)
        layout = QHBoxLayout()
        layout.addWidget(label_widget, Qt.AlignHCenter)
        layout.addWidget(input_widget, Qt.AlignHCenter)
        if input_parameter.info:
            info_button = InfoButton(input_parameter.info)
            layout.addWidget(info_button, Qt.AlignLeft)

        def get_input(*args):
            return input_widget.currentText()

        setattr(AbstractInputParametersWidget, input_parameter.name, property(get_input))
        return layout


from typing import List, Tuple
import logging


class Intervals(object):
    log = logging.getLogger(__name__)

    def __init__(self, size: int):
        self.fitted: List[int] = []
        self.size: int = size
        self.intervals: List[Tuple[int, int]] = []

    def in_interval(self, key: int) -> bool:
        for s, e in self.intervals:
            if s <= key <= e:
                return True
        return False

    def closest(self, key: int):
        min_key, min_distance = None, None
        for k in self.fitted:
            distance = abs(k - key)
            if min_distance is None or distance < min_distance:
                min_distance = distance
                min_key = k
        return min_key

    def add_interval(self, start: int = 0, end: int = -1):
        if end < 0:
            end = self.size + end
        if end <= start or start < 0 or end >= self.size:
            self.log.error(f'Wrong interval: [{start}:{end}].')
            return

        new_intervals = []
        for interval in self.intervals:
            if not (interval[0] > end or interval[1] < start):
                start = min(start, interval[0])
                end = max(end, interval[1])
            else:
                new_intervals.append(interval)
        new_intervals.append((start, end))
        self.intervals = sorted(new_intervals, key=lambda x: x[0])

    def add_key(self, key: int, bottom: bool = True):
        if key < 0 or key >= self.size:
            self.log.error(f'Wrong key {key}')
            return

        self.fitted.append(key)

        if bottom:
            for i, interval in enumerate(self.intervals):
                if interval[0] <= key <= interval[1]:
                    return
                if key < interval[0]:
                    self.intervals[i] = (key, interval[1])
                    return
            self.add_interval(key)
        else:
            for i, interval in enumerate(reversed(self.intervals)):
                if interval[0] <= key <= interval[1]:
                    return
                if key > interval[1]:
                    self.intervals[len(self.intervals) - i - 1] = (interval[0], key)
                    return
            self.add_interval(0, key)

    def del_key(self, key: int, bottom: bool = True):

        for i, interval in enumerate(self.intervals):
            if interval[0] <= key <= interval[1]:
                break
        else:
            return

        if bottom:
            self.intervals[i] = (interval[0], key - 1)
        else:
            self.intervals[i] = (key - 1, interval[1])

        if self.intervals[i][0] > self.intervals[i][1]:
            del self.intervals[i]

        try:
            self.fitted.remove(key)
        except ValueError:
            pass

    def __repr__(self):
        return repr(self.intervals)


if __name__ == '__main__':
    intervals = Intervals(100)
    print(intervals)
    intervals.add_key(0)
    print(intervals)
    intervals.del_key(10)
    print(intervals)
    intervals.add_key(20)
    print(intervals)
    intervals.add_key(15, False)
    print(intervals)
    print(intervals.fitted)



from typing import List, Tuple

import numpy as np

from .roi import Roi, RoiTypes


class RoiData(dict):
    def __init__(self, rois: List[Roi] = None):
        super().__init__()
        self._new_key = 0
        self._selected_keys = set()
        if rois:
            self.add_rois(rois)

    @property
    def selected_keys(self):
        yield from self._selected_keys

    @property
    def selected_num(self):
        return len(self._selected_keys)

    def select(self, key: int) -> Tuple:
        if len(self._selected_keys) == 1 and key in self._selected_keys:
            return ()
        try:
            for k in self._selected_keys:
                self[k].active = False
            self[key].active = True
            self._selected_keys.add(key)
            change_select = tuple(self._selected_keys)
            self._selected_keys = {key}
            return change_select

        except KeyError:
            raise ValueError(f'The selected roi {key} is missing.')

    def shift_select(self, key: int):
        try:
            self[key].active = not self[key].active
            if self[key].active:
                self._selected_keys.add(key)
            else:
                try:
                    self._selected_keys.remove(key)
                except KeyError:
                    pass
        except KeyError:
            raise ValueError(f'The selected roi {key} is missing.')

    @property
    def selected_rois(self) -> List[Roi]:
        return [self[i] for i in self._selected_keys]

    def select_all(self) -> List[int]:
        to_select = list()
        for roi in self.values():
            if not roi.active:
                roi.active = True
                to_select.append(roi.key)
        self._selected_keys = set(self.keys())
        return to_select

    def unselect_all(self) -> List[int]:
        to_unselect = list()
        for roi in self.values():
            if roi.active:
                roi.active = False
                to_unselect.append(roi.key)
        self._selected_keys = set()
        return to_unselect

    def add_roi(self, roi: Roi) -> None:
        if roi.key is None or roi.key < 0:
            roi.key = self._new_key
            self._new_key += 1
        self[roi.key] = roi

        if roi.active:
            self._selected_keys.add(roi.key)
        if not roi.name:
            roi.name = str(roi.key)

    def add_rois(self, roi_list: List[Roi]) -> List[int]:
        keys = list()
        to_select = list()
        for roi in roi_list:
            self.add_roi(roi)
            if roi.active:
                to_select.append(roi.key)
            keys.append(roi.key)
        return to_select

    def delete_roi(self, k: int):
        del self[k]

        if k in self._selected_keys:
            self._selected_keys.remove(k)

    def create_roi(self, radius: float, width: float, **params) -> Roi:
        roi = Roi(radius, width, **params)
        self.add_roi(roi)
        return roi

    def clear(self):
        super().clear()
        self._selected_keys = set()

    def change_ring_bounds(self, bounds: Tuple[float, float]) -> List[int]:
        angle, angle_std = bounds
        keys = list()
        for k, roi in self.items():
            if roi.type == RoiTypes.ring and (roi.angle != angle or roi.angle_std != angle_std):
                keys.append(k)
                roi.angle, roi.angle_std = bounds
        return keys

    def on_scale_changed(self, scale_change: float):
        for roi in self.values():
            roi.radius *= scale_change
            roi.width *= scale_change

    def apply_fit(self, rois: List[Roi]):
        keys_to_move = []
        rois_to_create = []

        for roi in rois:
            roi.active = True
            roi.movable = False

            if roi.key in self.keys():
                keys_to_move.append(roi.key)
                self[roi.key].update(roi)
                self.select(roi.key)
            else:
                rois_to_create.append(roi)

        self.add_rois(rois_to_create)
        keys_to_create = [roi.key for roi in rois_to_create]

        return tuple(keys_to_create), tuple(keys_to_move)

    def to_array(self) -> np.ndarray:
        return np.array([roi.to_array() for roi in self.values()])

    @classmethod
    def from_array(cls, arr: np.ndarray):
        return cls([Roi.from_array(a) for a in arr])



# -*- coding: utf-8 -*-
import sys
import traceback
import logging

from .tools import show_error

from PyQt5.QtCore import QObject, pyqtSignal


class UncaughtHook(QObject):

    _exception_caught = pyqtSignal(object)
    log = logging.getLogger(__name__)

    def __init__(self, *args, **kwargs):
        super(UncaughtHook, self).__init__(*args, **kwargs)

        # this registers the exception_hook() function as hook with the Python interpreter
        sys.excepthook = self.exception_hook

        # connect signal to execute the message box function always on main thread
        self._exception_caught.connect(show_error)

    def exception_hook(self, exc_type, exc_value, exc_traceback):
        """
        Function handling uncaught exceptions.
        It is triggered each time an uncaught exception occurs.
        """
        if issubclass(exc_type, KeyboardInterrupt):
            # ignore keyboard interrupt to support console applications
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
        else:
            exc_info = (exc_type, exc_value, exc_traceback)
            log_msg = '\n'.join([''.join(traceback.format_tb(exc_traceback)),
                                 '{0}: {1}'.format(exc_type.__name__, exc_value)])

            self.log.critical("Uncaught exception:\n {0}".format(log_msg), exc_info=exc_info)

            # trigger message box show
            self._exception_caught.emit(log_msg)




from typing import Tuple, NamedTuple

import cv2
import numpy as np

from .geometry import Geometry
from ..app.rois.roi import Roi

INTERPOLATION_ALGORITHMS = {
    'Nearest': cv2.INTER_NEAREST,
    'Bilinear': cv2.INTER_LINEAR,
    'Cubic': cv2.INTER_CUBIC,
    'Lanczos': cv2.INTER_LANCZOS4
}

INTERPOLATION_ALGORITHMS_INVERSED = {v: k for k, v in INTERPOLATION_ALGORITHMS.items()}


class InterpolationParams(NamedTuple):
    shape: Tuple[int, int] = (1024, 1024)
    algorithm: int = cv2.INTER_LINEAR


class PolarImage(object):

    def __init__(self, polar_img: np.ndarray = None,
                 parameters: InterpolationParams = None):
        self._polar_img = polar_img
        self._parameters = parameters or InterpolationParams()

    @property
    def polar_image(self) -> np.ndarray or None:
        return self._polar_img

    @property
    def polar_params(self) -> InterpolationParams:
        return self._parameters

    def set_params(self, shape: Tuple[int, int] = None, algorithm=None):
        shape = shape or self.polar_params.shape
        algorithm = algorithm if algorithm is not None else self.polar_params.algorithm

        self._parameters = InterpolationParams(shape, algorithm)

    def set_algorithm(self, algorithm: int, image: np.ndarray, geometry: Geometry):
        if self.polar_params.algorithm == algorithm:
            return

        self._parameters = InterpolationParams(self.polar_params.shape, algorithm)
        self.update(geometry, image)

    def update(self, geometry: Geometry, image: np.ndarray):
        if image is None:
            self._polar_img = None
            return

        yy, zz = geometry.polar_grids
        self._polar_img = self.calc_polar_image(image, yy, zz, self.polar_params.algorithm)

    def set_polar_image(self, img: np.ndarray):
        self._polar_img = img
        if img.shape != self._parameters.shape:
            self._parameters = InterpolationParams(img.shape, self._parameters.algorithm)

    @staticmethod
    def calc_polar_image(img: np.ndarray, yy: np.ndarray, zz: np.ndarray,
                         algorithm=cv2.INTER_LINEAR) -> np.ndarray or None:
        try:
            return cv2.remap(img.astype(np.float32),
                             yy.astype(np.float32),
                             zz.astype(np.float32),
                             interpolation=algorithm)
        except:
            return

    def get_radial_profile(self) -> np.ndarray or None:
        if self.polar_image is None:
            return
        return self.polar_image.sum(axis=0)

    def get_angular_profile(self, geometry: Geometry, roi: Roi) -> np.ndarray or None:
        if self.polar_image is None:
            return
        r1, r2 = roi.radius - roi.width / 2, roi.radius + roi.width / 2

        r_min, r_max = geometry.r_range
        scale = geometry.scale
        r_size = self.polar_image.shape[1]
        r_ratio = (r_max - r_min) / r_size * scale

        r1, r2 = int((r1 - r_min) / r_ratio), int((r2 - r_min) / r_ratio)
        r1, r2 = max(min((r1, r2)), 0), min(max((r1, r2)), r_size)

        if r1 > r_size or r2 < 0:
            return
        return self.polar_image[:, r1:r2].sum(axis=1)



# -*- coding: utf-8 -*-
import logging

from pyqtgraph import CircleROI
from PyQt5.QtWidgets import (QMainWindow, QWidget, QVBoxLayout, QPushButton)
from PyQt5.QtCore import pyqtSignal, Qt

from ..app.rois.roi import Roi
from ..app import App
from ..app.transformations import Transformation
from .basic_widgets import CustomImageViewer, LabeledSlider, BlackToolBar
from .roi_widgets.roi_2d_ring_widget import Roi2DRing
from .roi_widgets.abstract_roi_holder import AbstractRoiHolder
from .tools import Icon, center_widget

logger = logging.getLogger(__name__)


class GiwaxsImageViewer(AbstractRoiHolder, CustomImageViewer):
    class BeamCenterRoi(CircleROI):
        _ROI_SIZE = 1

        def __init__(self, beam_center, parent):
            CircleROI.__init__(self, (beam_center[1], beam_center[0]),
                               self._ROI_SIZE, movable=False, parent=parent)
            self._center = None
            self._scale = 1
            self.set_center(beam_center)

        def set_center(self, value: tuple, y=None, update=True, finish=True, ):
            self._center = value
            radius = self.size().x() / 2
            pos = (value[1] - radius, value[0] - radius)
            super(GiwaxsImageViewer.BeamCenterRoi, self).setPos(
                pos, y, update, finish)

        def set_size(self, size: float = None):
            size = size or self._ROI_SIZE
            size *= self._scale
            self.setSize((size, size), update=False, finish=False)
            self.set_center(self._center)

        def set_scale(self, scale: float):
            self._scale = scale
            self.set_size()

    def __init__(self, parent=None):
        AbstractRoiHolder.__init__(self, 'ImageViewer')
        CustomImageViewer.__init__(
            self, parent)
        self.image_plot.getAxis('bottom').setLabel(text='<math>Q<sub>xy</sub></math>', color='white')
        self.image_plot.getAxis('left').setLabel(text='<math>Q<sub>z</sub></math>', color='white')
        self.app = App()
        self._geometry_params_widget = None
        self.__init_center_roi()
        self.app.geometry_holder.sigScaleChanged.connect(self._on_scale_changed)
        self.app.geometry_holder.sigBeamCenterChanged.connect(self._on_beam_center_changed)
        self.app.image_holder.sigImageChanged.connect(self._on_image_changed)
        self.app.image_holder.sigEmptyImage.connect(self.clear_image)

    def _on_scale_changed(self):
        scale = self.app.geometry.scale
        self.set_scale(scale)
        self.center_roi.set_scale(scale)

    def _on_image_changed(self):
        image = self.app.image
        if image is not None:
            self.set_data(image)

    def _make_roi_widget(self, roi: Roi):
        roi_widget = Roi2DRing(roi)
        self.image_plot.addItem(roi_widget)
        return roi_widget

    def _delete_roi_widget(self, roi_widget: Roi2DRing):
        self.image_plot.removeItem(roi_widget)

    def __init_center_roi(self):
        beam_center = tuple(self.app.geometry.beam_center)
        self.center_roi = self.BeamCenterRoi(beam_center, parent=self.image_item)
        self.center_roi.setZValue(10)
        self.image_plot.addItem(self.center_roi)
        self.center_roi.hide()

        # self.angle_roi = self.ZeroAngleRoi(self.beam_center, 0, False, self.image_item)
        # self.angle_roi.setZValue(10)
        # self.image_plot.addItem(self.angle_roi)

    def open_geometry_parameters(self):
        image = self.app.image
        scale = self.app.geometry.scale
        beam_center = tuple(self.app.geometry.beam_center)
        if image is not None and self._geometry_params_widget is None:
            self._geometry_params_widget = GeometryParametersWidget(
                image.shape, beam_center, scale=scale)
            self.center_roi.show()
            self._geometry_params_widget.change_center.connect(lambda x:
                                                               self.app.geometry_holder.set_beam_center(x, False))
            self._geometry_params_widget.scale_changed.connect(
                self.app.geometry_holder.set_scale)
            self._geometry_params_widget.close_event.connect(self._on_closing_geometry_parameters)

    def _on_closing_geometry_parameters(self):
        self.center_roi.set_size()
        self._geometry_params_widget = None
        self.app.geometry_holder.sigGeometryChangeFinished.emit()

    def _on_beam_center_changed(self):
        beam_center = self.app.geometry.beam_center
        self.set_center((beam_center.y, beam_center.z), pixel_units=True)


class GeometryParametersWidget(QWidget):
    change_center = pyqtSignal(list)
    change_zero_angle = pyqtSignal(float)
    change_invert_angle = pyqtSignal(bool)
    scale_changed = pyqtSignal(float)

    close_event = pyqtSignal()

    def __init__(self, image_shape: tuple,
                 beam_center: tuple, zero_angle: float = 0,
                 angle_direction: bool = True, scale: float = 1):
        super(GeometryParametersWidget, self).__init__(None, Qt.WindowStaysOnTopHint)
        self.beam_center = list(beam_center)
        self.image_shape = image_shape
        self.zero_angle = zero_angle
        self.scale = scale
        self.angle_direction = angle_direction
        self._init__ui()
        self.setWindowTitle('Set geometry')
        self.setWindowIcon(Icon('setup'))
        center_widget(self)
        self.show()

    def closeEvent(self, a0) -> None:
        self.close_event.emit()
        QWidget.closeEvent(self, a0)

    def _init__ui(self):
        layout = QVBoxLayout(self)

        self.x_slider = LabeledSlider('Y center', (0, self.image_shape[1]),
                                      self.beam_center[1], self)
        self.x_slider.valueChanged.connect(self._connect_func(1))

        self.y_slider = LabeledSlider('Z center', (0, self.image_shape[0]),
                                      self.beam_center[0], self)
        self.y_slider.valueChanged.connect(self._connect_func(0))

        # self.angle_slider = AnimatedSlider('Zero angle', (0, 360),
        #                                    self.zero_angle, self,
        #                                    Qt.Horizontal, disable_changing_status=True)
        # self.angle_slider.valueChanged.connect(self._connect_func(2))
        #
        # self.invert_angle_box = QCheckBox('Invert angle')
        # self.invert_angle_box.toggled.connect(self._connect_func(3))

        self.scale_edit = LabeledSlider('Q to pixel ratio', (1e-10, 10),
                                        self.scale, self,
                                        decimals=5)
        self.scale_edit.valueChanged.connect(self.on_scale_changed)

        layout.addWidget(self.x_slider)
        layout.addWidget(self.y_slider)
        layout.addWidget(self.scale_edit)
        # layout.addWidget(self.angle_slider)
        # layout.addWidget(self.invert_angle_box)

    def on_scale_changed(self, value):
        self.scale = value
        self.scale_changed.emit(value)

    def _connect_func(self, ind: int):
        def beam_center_changed(value):
            self.beam_center[ind] = value
            self.change_center.emit(self.beam_center)

        def angle_zero_changed(value):
            self.zero_angle = value
            self.change_zero_angle.emit(self.zero_angle)

        def angle_direction_changed(value):
            self.angle_direction = value
            self.change_invert_angle.emit(self.angle_direction)

        if ind < 2:
            return beam_center_changed
        elif ind == 2:
            return angle_zero_changed
        else:
            return angle_direction_changed


class Basic2DImageWidget(QMainWindow):

    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)
        self.app = App()
        self.image_viewer = GiwaxsImageViewer(self)
        self.setCentralWidget(self.image_viewer)
        self._init_toolbar()

    def _init_toolbar(self):
        toolbar = BlackToolBar('Geometry', self)
        self.addToolBar(toolbar)

        rotate_action = toolbar.addAction(Icon('rotate'), 'Rotate')
        rotate_action.triggered.connect(
            lambda: self.app.geometry_holder.add_transform(Transformation.rotate_right))

        flip_h = toolbar.addAction(Icon('flip_horizontal'), 'Horizontal flip')
        flip_h.triggered.connect(
            lambda: self.app.geometry_holder.add_transform(Transformation.horizontal_flip))

        flip_v = toolbar.addAction(Icon('flip_vertical'), 'Vertical flip')
        flip_v.triggered.connect(
            lambda: self.app.geometry_holder.add_transform(Transformation.vertical_flip))

        set_beam_center_action = toolbar.addAction(Icon('center'), 'Beam center')
        set_beam_center_action.triggered.connect(
            self.image_viewer.open_geometry_parameters)

        self._set_default_geometry_button = QPushButton('Save as default geometry')
        self._set_default_geometry_button.clicked.connect(self.app.geometry_holder.save_as_default)
        toolbar.addWidget(self._set_default_geometry_button)



import logging
from typing import Dict, Tuple, List, DefaultDict, Set
from collections import defaultdict
import gc

from PyQt5.QtCore import QObject, pyqtSignal, Qt, pyqtSlot

from PyQt5.QtWidgets import (QPlainTextEdit, QTreeWidget,
                             QTreeWidgetItem, QWidget, QPushButton,
                             QVBoxLayout, QApplication, QLabel, QSplitter,
                             QListWidget, QListWidgetItem, QLineEdit, QMenu,
                             QCheckBox, QGridLayout)

from PyQt5.QtGui import QColor

from ..app import App
from ..app.debug_tracker import ObjectTracker, ObjectStatus


class QTextEditLogger(logging.Handler, QObject):
    appendPlainText = pyqtSignal(str)

    def __init__(self, parent):
        super().__init__()
        QObject.__init__(self)
        self.widget = QPlainTextEdit(parent)
        self.widget.setReadOnly(True)
        self.appendPlainText.connect(self.widget.appendPlainText)
        self.setFormatter(
            logging.Formatter(
                '%(asctime)s %(levelname)s %(module)s %(funcName)s %(message)s'))
        logging.getLogger().addHandler(self)
        logging.getLogger().setLevel(logging.DEBUG)

    def emit(self, record):
        msg = self.format(record)
        try:
            self.appendPlainText.emit(msg)
        except RuntimeError:
            pass


class TrackerItem(QListWidgetItem):
    def __init__(self, parent: QListWidget, obj: ObjectTracker):
        super().__init__(obj.name, parent)
        self.id = obj.id


class TrackerWidget(QWidget):
    log = logging.getLogger(__name__)

    def __init__(self, parent):
        super().__init__(parent=parent)
        self.tracker = App().debug_tracker
        self.tracker.signals.sigObjectAdded.connect(self._add_object)
        self.tracker.signals.sigStatusUpdated.connect(self._status_updated)
        self.item_dict: Dict[str, TrackerItem] = {}
        self._init_ui()

    def _init_ui(self):
        layout = QGridLayout(self)

        self.update_button = QPushButton('Update', self)
        self.clear_button = QPushButton('Clear', self)
        self.collect_checkbox = QCheckBox('GC collect', self)
        self.search_line = QLineEdit(parent=self)
        self.splitter = QSplitter(orientation=Qt.Horizontal, parent=self)

        self.list_exists, w1 = self._get_list('Exist')
        self.list_deleted_by_c, w2 = self._get_list('Deleted by c')
        self.list_deleted_by_python, w3 = self._get_list('Deleted by python')
        self.list_safely_deleted, w4 = self._get_list('Safely deleted')

        self.splitter.addWidget(w1)
        self.splitter.addWidget(w2)
        self.splitter.addWidget(w3)
        self.splitter.addWidget(w4)

        num = 3

        layout.addWidget(self.update_button, 0, 0, 1, num)
        layout.addWidget(self.clear_button, 1, 0, 1, num)
        layout.addWidget(self.collect_checkbox, 2, 0)
        layout.addWidget(QLabel('Search: '), 2, 1)
        layout.addWidget(self.search_line, 2, 2)
        layout.addWidget(self.splitter, 3, 0, 1, num)

        self.update_button.clicked.connect(self._update)
        self.clear_button.clicked.connect(self._clear)
        self.search_line.editingFinished.connect(self._search)

        self.list_exists.setContextMenuPolicy(Qt.CustomContextMenu)
        self.list_exists.customContextMenuRequested.connect(self._custom_menu(self.list_exists))

        self.list_deleted_by_c.setContextMenuPolicy(Qt.CustomContextMenu)
        self.list_deleted_by_c.customContextMenuRequested.connect(self._custom_menu(self.list_deleted_by_c))

    def _custom_menu(self, widget):
        def menu(pos):
            item = widget.itemAt(pos)
            if not item:
                return

            q_menu = QMenu()
            q_menu.addAction('Referrers', lambda *args, key=item.id: self._log_referrers(key))
            q_menu.addAction('Is tracked', lambda *args, key=item.id: self._is_tracked(key))
            q_menu.addAction('Remove', lambda *args, key=item.id: self._remove(key))
            self.tracker.add_object(q_menu)
            q_menu.exec(widget.mapToGlobal(pos))
        return menu

    @pyqtSlot(name='searchItems')
    def _search(self):
        text = self.search_line.text()
        for item in self.item_dict.values():
            if text in item.text():
                item.setHidden(False)
            else:
                item.setHidden(True)

    @pyqtSlot(name='updateLists')
    def _update(self):
        if self.collect_checkbox.isChecked():
            gc.collect()
        self.tracker.update()

    def _log_referrers(self, key: str):
        try:
            obj = self.tracker.objects[key]
        except KeyError as err:
            self.log.exception(err)
            self._remove(key)
            return

        referrers = obj.get_referrers()
        if not referrers:
            self.log.error('No referrers found!')
        else:
            self.log.debug(f'Referrers to {obj.name}:')
            for r in referrers:
                self.log.debug(f'{r.__class__.__name__}: {r}')
            self.log.debug(f'--------------------')

    def _is_tracked(self, key: str):
        try:
            obj = self.tracker.objects[key]
        except KeyError as err:
            self.log.exception(err)
            return
        if obj.is_tracked():
            self.log.debug(f'{obj.name} is tracked by gc.')
        else:
            self.log.error(f'{obj.name} is NOT tracked by gc!')

    def _remove(self, key: str):
        try:
            self.tracker.objects.pop(key)
        except KeyError:
            self.log.debug(f'{key} not in tracker')
        try:
            item = self.item_dict.pop(key)
            parent = item.listWidget()
            parent.takeItem(parent.row(item))
        except KeyError:
            self.log.debug(f'{key} not in item_dict')
            return

    @pyqtSlot(list, name='statusUpdated')
    def _status_updated(self, key_list: list):
        for key in key_list:
            status = self.tracker.objects[key].status
            item = self.item_dict[key]
            parent = item.listWidget()
            new_parent = self._get_list_by_status(status)
            new_parent.insertItem(new_parent.count(), parent.takeItem(parent.row(item)))

    def _clear(self):
        to_delete = self.tracker.clear_deleted()
        for key in to_delete:
            item = self.item_dict.pop(key)
            item.listWidget().removeItemWidget(item)

        self.list_safely_deleted.clear()

    def _get_list_by_status(self, status: ObjectStatus):
        if status == ObjectStatus.EXISTS:
            return self.list_exists
        if status == ObjectStatus.SAFELY_DELETED:
            return self.list_safely_deleted
        if status == ObjectStatus.DELETED_BY_C:
            return self.list_deleted_by_c
        if status == ObjectStatus.DELETED_BY_PYTHON:
            return self.list_deleted_by_python
        raise ValueError('Unknown status')

    @pyqtSlot(str, name='objectAdded')
    def _add_object(self, key: str):
        self.item_dict[key] = TrackerItem(self.list_exists, self.tracker.objects[key])

    def _get_list(self, name):
        widget = QWidget(self)
        label = QLabel(name, widget)
        list_widget = QListWidget(widget)

        layout = QVBoxLayout(widget)
        layout.addWidget(label)
        layout.addWidget(list_widget)

        return list_widget, widget


class TreeWidgetItem(QTreeWidgetItem):
    def __lt__(self, other):
        return int(self.text(2)) < int(other.text(2))


def get_color(fr: float):
    return QColor(0, 255, 0, int(fr * 255))


class WidgetList(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.id_dict: Dict[int, Tuple[List[str], int]] = {}
        self.item_dict: DefaultDict[str, Set[tuple]] = defaultdict(set)
        self._init_ui()

    def _init_ui(self):
        layout = QVBoxLayout(self)
        self.splitter = QSplitter(orientation=Qt.Vertical, parent=self)

        self.label = QLabel('Number of widgets (not calculated)')
        self.update_button = QPushButton('Update')
        self.widget_list = QTreeWidget(self)
        self.differ_list = QListWidget(self)

        self.search_widget = QWidget(self)
        self.search_line = QLineEdit(self.search_widget)
        self.search_tree = QTreeWidget(self.search_widget)

        self.widget_list.setHeaderLabels(['Class name', 'Object name', 'Number of children', 'id'])
        self.search_tree.setHeaderLabels(['Class name', 'Object name', 'Number of children', 'id'])

        s_layout = QVBoxLayout(self.search_widget)
        s_layout.addWidget(self.search_line)
        s_layout.addWidget(self.search_tree)

        layout.addWidget(self.label)
        layout.addWidget(self.update_button)
        layout.addWidget(self.splitter)

        self.splitter.addWidget(self.widget_list)
        self.splitter.addWidget(self.differ_list)
        self.splitter.addWidget(self.search_widget)

        self.update_button.clicked.connect(self.update_widget_list)
        self.search_line.editingFinished.connect(self._update_search_list)

        self.differ_list.hide()

    @pyqtSlot(name='updateSearchList')
    def _update_search_list(self):
        search: str = self.search_line.text().lower()
        self.search_tree.clear()
        for key, value in self.item_dict.items():
            if key.lower().startswith(search):
                for name_list in value:
                    TreeWidgetItem(self.search_tree, name_list)

    def update_widget_list(self):
        self.widget_list.clear()
        self.differ_list.clear()
        self.item_dict.clear()

        count = 0
        new_id_dict: Dict[int, Tuple[list, int]] = {}

        for obj in QApplication.topLevelWidgets():
            count += self._update_list(obj, self.widget_list, new_id_dict)

        self.widget_list.update()
        self.label.setText(f'Number of widgets = {count}')
        self.widget_list.sortByColumn(2, Qt.DescendingOrder)

        for value in self.id_dict.values():
            QListWidgetItem(' '.join(value[0]), self.differ_list)

        self.id_dict = new_id_dict

    def _update_list(self, parent_obj: QObject,
                     parent_node: QTreeWidget or QTreeWidgetItem, new_id_dict: Dict[int, Tuple[List[str], int]]):
        count = 0

        for obj in parent_obj.children():
            name_list = self._get_name_list(obj)
            item = TreeWidgetItem(parent_node, name_list)

            obj_count = 0

            for o in obj.children():
                obj_count += self._update_list(o, item, new_id_dict)

            item.setText(2, str(obj_count))

            name_list[2] = str(obj_count)
            self.item_dict[str(obj.__class__.__name__)].add(tuple(name_list))

            id_ = id(obj)
            new_id_dict[id_] = [item.text(i) for i in range(4)], obj_count

            _, prev_count = self.id_dict.pop(id_, (None, None))

            if prev_count is None:
                fr = 1
            elif prev_count + obj_count == 0:
                fr = 0
            else:
                fr = abs(prev_count - obj_count) / (prev_count + obj_count)
            item.setBackground(3, get_color(fr))

            count += 1 + obj_count

        return count

    @staticmethod
    def _get_name_list(obj: QObject):
        return [str(obj.__class__.__name__), obj.objectName(), '', str(id(obj))]


class DebugWindow(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._init_ui()
        self.show()

    def _init_ui(self):
        layout = QVBoxLayout(self)
        self.setWindowFlag(Qt.Window)
        self.setWindowTitle('Debugging Window')

        self.logging_widget = QTextEditLogger(self)
        # self.widget_list = WidgetList(self)
        self.widget_list = TrackerWidget(self)
        self.splitter = QSplitter(orientation=Qt.Vertical, parent=self)
        self.splitter.addWidget(self.logging_widget.widget)
        self.splitter.addWidget(self.widget_list)

        layout.addWidget(self.splitter)



from abc import abstractmethod
from typing import Dict

from .abstract_roi_widget import AbstractRoiWidget
from ...app.app import App
from ...app.rois.roi import Roi


class AbstractRoiHolder(object):
    def __init__(self, name: str):
        self._name = name
        self._roi_dict = App().roi_dict
        self._roi_widgets: Dict[int, AbstractRoiWidget] = {}
        self._init_connect()

    def _init_connect(self):
        self._roi_dict.sig_roi_created.connect(self._create_roi)
        self._roi_dict.sig_roi_deleted.connect(self._delete_roi)
        self._roi_dict.sig_roi_moved.connect(self._move_rois)
        self._roi_dict.sig_selected.connect(self._update_select)
        self._roi_dict.sig_fixed.connect(self._fix_rois)
        self._roi_dict.sig_unfixed.connect(self._unfix_rois)

    def _create_roi(self, keys: tuple):
        for key in keys:
            self._roi_widgets[key] = self._make_roi_widget(self._roi_dict[key])
            self._connect_roi_widget(self._roi_widgets[key])

    def _move_rois(self, keys: tuple, name: str):
        if name != self._name:
            for key in keys:
                try:
                    self._roi_widgets[key].move_roi()
                except KeyError:
                    pass

    def _connect_roi_widget(self, roi_widget: AbstractRoiWidget):
        roi_widget.sigRoiMoved.connect(
            lambda k: self._roi_dict.move_roi(k, self._name))
        roi_widget.sigSelected.connect(self._roi_dict.select)
        roi_widget.sigShiftSelected.connect(self._roi_dict.shift_select)

    def _update_select(self, keys: tuple):
        for key in keys:
            self._roi_widgets[key].update_select()

    def _delete_roi(self, keys: tuple):
        for key in keys:
            try:
                self._delete_roi_widget(self._roi_widgets.pop(key))
            except KeyError:
                pass

    def _fix_rois(self, keys: tuple):
        for key in keys:
            try:
                self._roi_widgets[key].fix()
            except KeyError:
                pass

    def _unfix_rois(self, keys: tuple):
        for key in keys:
            try:
                self._roi_widgets[key].unfix()
            except KeyError:
                pass

    @abstractmethod
    def _delete_roi_widget(self, roi_widget) -> None:
        pass

    @abstractmethod
    def _make_roi_widget(self, roi: Roi) -> AbstractRoiWidget:
        pass
